import collections.abc as collections_abc
import inspect
import typing
import warnings
from abc import ABC, abstractmethod
from collections import OrderedDict, defaultdict
from dataclasses import dataclass, fields, is_dataclass
from datetime import date, datetime
from enum import Enum, IntEnum
from types import GenericAlias as GenericAlias_
from types import UnionType
from typing import Any, Dict, Iterable, Sequence, Type, TypeAlias, Union
from typing import _AnnotatedAlias as AnnotatedAlias
from typing import _GenericAlias, get_type_hints
from uuid import UUID

from guardpost import AuthenticationHandler
from guardpost.common import AuthenticatedRequirement
from openapidocs.common import Format, Serializer
from openapidocs.v3 import (
    APIKeySecurity,
    Components,
    Example,
    Header,
    HTTPSecurity,
    Info,
    MediaType,
    OpenAPI,
    Operation,
    Parameter,
    ParameterLocation,
    PathItem,
    Reference,
    RequestBody,
)
from openapidocs.v3 import Response as ResponseDoc
from openapidocs.v3 import (
    Schema,
    Security,
    SecurityRequirement,
    SecurityScheme,
    Server,
    Tag,
    ValueFormat,
    ValueType,
)

from blacksheep.server.authentication.apikey import APIKeyAuthentication, APIKeyLocation
from blacksheep.server.authentication.basic import BasicAuthentication
from blacksheep.server.bindings import (
    Binder,
    BodyBinder,
    CookieBinder,
    HeaderBinder,
    QueryBinder,
    RouteBinder,
    empty,
)
from blacksheep.server.openapi.docstrings import (
    DocstringInfo,
    get_handler_docstring_info,
)
from blacksheep.server.openapi.exceptions import DuplicatedContentTypeDocsException
from blacksheep.server.routing import Router

from ..application import Application
from .common import (
    APIDocsHandler,
    ContentInfo,
    DirectSchema,
    EndpointDocs,
    HeaderInfo,
    ParameterInfo,
    ParameterSource,
    RequestBodyInfo,
    ResponseExample,
    ResponseInfo,
    ResponseStatusType,
    response_status_to_str,
)

try:
    # JWT dependencies are optional (cryptography, PyJWT)
    from blacksheep.server.authentication.jwt import JWTBearerAuthentication
except ImportError:  # pragma: no cover

    class JWTBearerAuthentication:
        pass


try:
    from pydantic import BaseModel
except ImportError:  # pragma: no cover
    # noqa
    BaseModel = ...


# Pydantic v2 specific
try:
    from pydantic import TypeAdapter
    from pydantic.dataclasses import is_pydantic_dataclass
except ImportError:  # pragma: no cover
    # noqa
    TypeAdapter = ...
    is_pydantic_dataclass = ...


# Alias to support both typing List[T]/list[T] as GenericAlias
GenericAlias: TypeAlias = _GenericAlias | GenericAlias_


def _is_union_type(annotation):
    if isinstance(annotation, UnionType):  # type: ignore
        return True
    return False


# endregion


def get_origin(object_type):
    return typing.get_origin(object_type)


def check_union(object_type: Any) -> tuple[bool, Any]:
    if (
        hasattr(object_type, "__origin__")
        and object_type.__origin__ is Union
        or _is_union_type(object_type)
        or (hasattr(object_type, "nullable"))
        or (hasattr(object_type, "is_required"))
    ):
        if hasattr(object_type, "nullable"):
            # Pydantic v1
            return object_type.nullable, object_type
        if hasattr(object_type, "is_required"):
            # Pydantic v2
            return object_type.is_required(), object_type

        if type(None) in object_type.__args__ and len(object_type.__args__) == 2:
            for possible_type in object_type.__args__:
                if type(None) is possible_type:
                    continue
                return True, possible_type
        return type(None) not in object_type.__args__, object_type
    return False, object_type


def is_ignored_parameter(param_name: str, matching_binder: Binder | None) -> bool:
    # if a binder is used, handle only those that can be mapped to a type of
    # OpenAPI Documentation parameter's location
    if matching_binder and not isinstance(
        matching_binder,
        (
            QueryBinder,
            RouteBinder,
            HeaderBinder,
            CookieBinder,
        ),
    ):
        return True

    return param_name == "request" or param_name == "services"


@dataclass
class FieldInfo:
    """
    This class is a data structure used to represent information about a field in an
    object type (e.g., a dataclass) when generating OpenAPI documentation.
    Specifically, it encapsulates the following details about a field:
        name: The name of the field.
        type: The type of the field, which can be either a
              Python type (e.g., str, int) or an OpenAPI Schema object.
        schema: The dictionary obtained from Pydantic's schema generation.

    This class is used by the ObjectTypeHandler implementations
    to extract and organize field information from object types.
    """

    name: str
    type: Type | Schema
    schema: dict[str, Any] | None = None  # New property to store Pydantic schema


class ObjectTypeHandler(ABC):
    """
    Abstract class describing the interface used to generate an OpenAPI Documentation
    schema for an object type.
    """

    @abstractmethod
    def handles_type(self, object_type) -> bool:
        """
        Returns a value indicating whether the given type is handled by this
        object type handler.
        """

    @abstractmethod
    def set_type_schema(self, object_type, context: "OpenAPIHandler") -> None:
        """
        Allow to set the schema for the type as a whole, skipping inspections of its
        fields. This is to enable delegating the generation of type schema to external
        libraries and better support Pydantic and other libraries that can generate the
        schema.
        """

    @abstractmethod
    def get_type_fields(self, object_type, register_type) -> list[FieldInfo]:
        """
        Returns a set of fields to be used for the handled type.
        """


class SecuritySchemeHandler(ABC):
    """
    Abstract class describing the interface used to generate an OpenAPI Documentation
    security scheme for an authentication handler.
    """

    @abstractmethod
    def get_security_schemes(
        self, handler: AuthenticationHandler
    ) -> Iterable[tuple[str, SecurityScheme, SecurityRequirement]]:
        """
        Returns an iterable of tuples containing:
            - the name of the security scheme,
            - the security scheme definition itself.
        """


class APIKeySecuritySchemeHandler(SecuritySchemeHandler):
    """
    A SecuritySchemeHandler that can handle APIKeyAuthentication handlers.
    """

    @staticmethod
    def _get_api_key_location(location: APIKeyLocation) -> ParameterLocation:
        if location == APIKeyLocation.HEADER:
            return ParameterLocation.HEADER
        if location == APIKeyLocation.QUERY:
            return ParameterLocation.QUERY
        if location == APIKeyLocation.COOKIE:
            return ParameterLocation.COOKIE
        raise ValueError(f"Unknown APIKeyLocation: {location}")

    def get_security_schemes(
        self, handler: AuthenticationHandler
    ) -> Iterable[tuple[str, SecurityScheme, SecurityRequirement]]:
        if isinstance(handler, APIKeyAuthentication):
            yield handler.scheme, APIKeySecurity(
                name=handler.param_name,
                in_=self._get_api_key_location(handler.location),
                description=handler.description,
            ), SecurityRequirement(handler.scheme, [])


class BasicAuthenticationSecuritySchemeHandler(SecuritySchemeHandler):
    """
    A SecuritySchemeHandler that can handle APIKeyAuthentication handlers.
    """

    def get_security_schemes(
        self, handler: AuthenticationHandler
    ) -> Iterable[tuple[str, SecurityScheme, SecurityRequirement]]:
        if isinstance(handler, BasicAuthentication):
            yield handler.scheme, HTTPSecurity(
                scheme="basic",
                description=handler.description,
            ), SecurityRequirement(handler.scheme, [])


class JWTAuthenticationSecuritySchemeHandler(SecuritySchemeHandler):
    """
    A SecuritySchemeHandler that can handle APIKeyAuthentication handlers.
    """

    def get_security_schemes(
        self, handler: AuthenticationHandler
    ) -> Iterable[tuple[str, SecurityScheme, SecurityRequirement]]:
        if isinstance(handler, JWTBearerAuthentication):
            yield handler.scheme, HTTPSecurity(
                scheme="bearer", bearer_format="JWT"
            ), SecurityRequirement(handler.scheme, [])


def _try_is_subclass(object_type, check_type):
    try:
        return issubclass(object_type, check_type)
    except TypeError:
        return False


class DataClassTypeHandler(ObjectTypeHandler):
    """
    An ObjectTypeHandler that can handle built-in dataclasses.
    """

    def handles_type(self, object_type) -> bool:
        return is_dataclass(object_type)

    def set_type_schema(self, object_type, context: "OpenAPIHandler") -> None:
        # Raise not implemented, because fields inspection should run for this handler.
        raise NotImplementedError()

    def get_type_fields(self, object_type, register_type) -> list[FieldInfo]:
        return [FieldInfo(field.name, field.type) for field in fields(object_type)]


class PydanticModelTypeHandler(ObjectTypeHandler):
    """
    An ObjectTypeHandler that can handle subclasses of Pydantic BaseModel.
    """

    def handles_type(self, object_type) -> bool:
        if isinstance(object_type, GenericAlias) or isinstance(object_type, UnionType):
            # support for Generic BaseModel here is better since it can extract items
            # type, skip this;
            # anyway using Generic with BaseModel doesn't seem to be recommended by
            # pydantic:
            # https://pydantic-docs.helpmanual.io/usage/types/#generic-classes-as-types
            return False
        if (
            BaseModel is not ...
            and inspect.isclass(object_type)
            and _try_is_subclass(object_type, BaseModel)  # type: ignore
        ):
            return True
        return False

    def set_type_schema(self, object_type, context: "OpenAPIHandler") -> None:
        """
        Fully delegates to Pydantic the generation of the schema for the class, however
        applies some changes to replace $defs with #/components/schemas for consistency.
        """
        schema = self._get_object_schema(object_type)
        self._normalize_dict_schema(schema, context)
        context.set_type_schema(object_type, schema)

    def get_type_fields(self, object_type, register_type) -> list[FieldInfo]:
        raise NotImplementedError()

    def _get_object_schema(self, object_type):
        if hasattr(object_type, "model_json_schema"):
            # Pydantic v2
            return object_type.model_json_schema()
        else:
            # Pydantic v1
            return object_type.schema()

    def _normalize_dict_schema(self, schema, context):
        """
        Apply definitions in $defs to components.schemas, and replace references defined
        as #/$defs/Name with #/components/schemas/Name.
        """

        self._normalize_dict_schema_apply_defs(schema, context)
        self._normalize_dict_schema_replace_refs(schema)

    def _get_defs_key(self) -> str:
        if TypeAdapter is ...:
            # Pydantic v1
            return "definitions"
        # Pydantic v2
        return "$defs"

    def _normalize_dict_schema_apply_defs(self, schema, context: "OpenAPIHandler"):
        definitions_keys = self._get_defs_key()

        if definitions_keys in schema:
            components = context.components
            if components.schemas is None:
                components.schemas = {}
            for key, value in schema[definitions_keys].items():
                self._normalize_dict_schema_replace_refs(value)
                components.schemas[key] = DirectSchema(value)  # type: ignore
            del schema[definitions_keys]

    def _normalize_dict_schema_replace_refs(self, schema):
        """
        Replace references defined as #/$defs/Name or #/definitions/Name
        with #/components/schemas/Name.
        """
        definitions_keys = self._get_defs_key()
        def_ref = f"#/{definitions_keys}/"
        to_replace = None

        for key, value in schema.items():
            if isinstance(value, dict):
                self._normalize_dict_schema_replace_refs(value)
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        self._normalize_dict_schema_replace_refs(item)
            if key == "$ref" and isinstance(value, str) and value.startswith(def_ref):
                to_replace = value
        if to_replace:
            schema["$ref"] = "#/components/schemas/" + to_replace.removeprefix(def_ref)


class PydanticDataClassTypeHandler(PydanticModelTypeHandler):
    """
    An ObjectTypeHandler that can handle Pydantic dataclasses.
    """

    def handles_type(self, object_type) -> bool:
        return is_pydantic_dataclass is not ... and is_pydantic_dataclass(object_type)

    def _get_object_schema(self, object_type):
        if TypeAdapter is ...:
            raise TypeError("Missing Pydantic")
        return TypeAdapter(object_type).json_schema()


class OpenAPIHandler(APIDocsHandler[OpenAPI]):
    """
    Handles the automatic generation of OpenAPI Documentation, specification v3
    for a web application, exposing methods to enrich the documentation with details
    through decorators.
    """

    def __init__(
        self,
        *,
        info: Info,
        ui_path: str = "/docs",
        json_spec_path: str = "openapi.json",
        yaml_spec_path: str = "openapi.yaml",
        preferred_format: Format = Format.JSON,
        anonymous_access: bool = True,
        tags: list[Tag] | None = None,
        security_schemes: dict[str, SecurityScheme] | None = None,
        servers: Sequence[Server] | None = None,
        serializer: Serializer | None = None,
    ) -> None:
        super().__init__(
            ui_path=ui_path,
            json_spec_path=json_spec_path,
            yaml_spec_path=yaml_spec_path,
            preferred_format=preferred_format,
            anonymous_access=anonymous_access,
            serializer=serializer,
        )
        self.info = info
        self._tags = tags
        self.components = Components(security_schemes=security_schemes)  # type: ignore
        self._objects_references: dict[Any, Reference] = {}
        self.servers: list[Server] = list(servers) if servers else []
        self.common_responses: dict[ResponseStatusType, ResponseDoc] = {}
        self._object_types_handlers: list[ObjectTypeHandler] = [
            DataClassTypeHandler(),
            PydanticModelTypeHandler(),
            PydanticDataClassTypeHandler(),
        ]
        self._security_schemes_handlers: list[SecuritySchemeHandler] = [
            APIKeySecuritySchemeHandler(),
            BasicAuthenticationSecuritySchemeHandler(),
            JWTAuthenticationSecuritySchemeHandler(),
        ]
        self._binder_docs: dict[Type[Binder], Iterable[Parameter | Reference]] = {}
        self.security: Security | None = None

    @property
    def object_types_handlers(self) -> list[ObjectTypeHandler]:
        """
        Gets the list of objects responsible of generating OpenAPI Documentation
        schemas for object types. Extend this list with custom objects to control how
        OpenAPI Schemas are handled for specific types.
        """
        return self._object_types_handlers

    @property
    def security_schemes_handlers(self) -> list[SecuritySchemeHandler]:
        """
        Gets the list of objects responsible of generating OpenAPI Documentation
        for security schemas. These objects are used to generate security schemes
        from the application's authentication strategy.
        """
        return self._security_schemes_handlers

    def get_ui_page_title(self) -> str:
        return self.info.title

    def _iter_tags(self, paths: dict[str, PathItem]) -> Iterable[str]:
        methods = (
            "get",
            "post",
            "put",
            "delete",
            "options",
            "head",
            "patch",
            "trace",
        )

        for path in paths.values():
            for method in methods:
                prop: Operation | None = getattr(path, method)

                if prop is not None and prop.tags is not None:
                    yield from prop.tags

    def _tags_from_paths(self, paths: dict[str, PathItem]) -> list[Tag]:
        unique_tags = set(self._iter_tags(paths))
        return [Tag(name=name) for name in sorted(unique_tags)]

    def generate_documentation(self, app: Application) -> OpenAPI:
        self._optimize_binders_docs()
        self._generate_security_schemes(app)
        paths = self.get_paths(app)
        return OpenAPI(
            info=self.info,
            paths=paths,
            components=self.components,
            tags=self._tags or self._tags_from_paths(paths),
            json_schema_dialect=None,  # type: ignore
            security=self.security,
        )

    def get_paths(self, app: Application, path_prefix: str = "") -> dict[str, PathItem]:
        own_paths = self.get_routes_docs(app.router, path_prefix)

        if app.mount_registry.mounted_apps and app.mount_registry.handle_docs:
            for route in app.mount_registry.mounted_apps:
                if isinstance(route.handler, Application):
                    full_prefix = route.pattern.decode().rstrip("/*")
                    if path_prefix:
                        full_prefix = path_prefix.rstrip("/") + full_prefix

                    child_docs = self.get_paths(
                        route.handler,
                        full_prefix,
                    )
                    own_paths.update(child_docs)

        return own_paths

    def get_type_name_for_annotated(
        self,
        object_type: AnnotatedAlias,
        context_type_args: dict[Any, Type] | None = None,
    ) -> str:
        """
        This method returns a type name for an annotated type.
        """
        assert isinstance(
            object_type, AnnotatedAlias
        ), "This method requires an annotated type"
        # Note: by default returns a string respectful of this requirement:
        # $ref values must be RFC3986-compliant percent-encoded URIs
        # Therefore, a generic that would be expressed in Python: Example[Foo, Bar]
        # and C# or TypeScript Example<Foo, Bar>
        # Becomes here represented as: ExampleOfFooAndBar
        # For typing.Annotated[T, ...], the underlying annotated type T is the
        # first element of typing.get_args(object_type). Using get_origin here
        # would return 'Annotated' itself; we want the actual base type.
        args = typing.get_args(object_type)
        origin = args[0] if args else get_origin(object_type)
        annotations = object_type.__metadata__
        annotations_repr = "And".join(
            self.get_type_name(annotation, context_type_args)
            for annotation in annotations
        )
        return f"{self.get_type_name(origin)}Of{annotations_repr}"

    def get_type_name_for_generic(
        self,
        object_type: GenericAlias,
        context_type_args: dict[Any, Type] | None = None,
    ) -> str:
        """
        This method returns a type name for a generic type.
        """
        assert isinstance(object_type, GenericAlias) or isinstance(
            object_type, UnionType
        ), "This method requires a generic"
        # Note: by default returns a string respectful of this requirement:
        # $ref values must be RFC3986-compliant percent-encoded URIs
        # Therefore, a generic that would be expressed in Python: Example[Foo, Bar]
        # and C# or TypeScript Example<Foo, Bar>
        # Becomes here represented as: ExampleOfFooAndBar
        origin = get_origin(object_type)
        args = typing.get_args(object_type)
        args_repr = "And".join(
            self.get_type_name(arg, context_type_args) for arg in args
        )
        return f"{self.get_type_name(origin)}Of{args_repr}"

    def get_type_name(
        self, object_type, context_type_args: dict[Any, Type] | None = None
    ) -> str:
        if context_type_args and object_type in context_type_args:
            object_type = context_type_args.get(object_type)
        if isinstance(object_type, AnnotatedAlias):
            return self.get_type_name_for_annotated(object_type, context_type_args)
        if isinstance(object_type, GenericAlias) or isinstance(object_type, UnionType):
            return self.get_type_name_for_generic(object_type, context_type_args)
        # Workaround for built-in collection types in Python 3.9+ to have them
        # capitalized in schema names
        if object_type in (list, tuple, set, dict) and hasattr(object_type, "__name__"):
            return object_type.__name__.capitalize()
        if hasattr(object_type, "__name__"):
            return object_type.__name__
        if object_type is Union:
            # Python 3.9 and 3.8 would fall here, Union type has a "_name" property but
            # no "__name__"
            return "Union"
        raise ValueError(
            f"Cannot obtain a name for object_type parameter: {object_type!r}"
        )

    def app_requires_auth(self, app: Application) -> bool:
        """
        Returns a value indicating whether the application **requires** authentication
        for all endpoints by default. If it does, the generated OpenAPI Documentation
        does not include an empty object in the global "security" section.
        Override this method in subclasses for finer control on this feature.
        """
        authorization_strategy = app.authorization_strategy
        if (
            authorization_strategy is None
            or authorization_strategy.default_policy is None
        ):
            return False
        default_reqs = authorization_strategy.default_policy.requirements

        if AuthenticatedRequirement in default_reqs or any(
            isinstance(req, AuthenticatedRequirement) for req in default_reqs
        ):
            return True
        return False

    def _generate_security_schemes(self, app: Application):
        # If the user did not specify security schemes, tries to generate security
        # schemes from the configured authentication handlers.
        # Important:
        schemes = self.components.security_schemes or {}

        if not schemes and app.authentication_strategy is not None:
            # Note: auth_handler can be defined as a type! Because Guardpost supports
            # dependency injection... We need to resolve services at application start
            # to generate OpenAPI Docs.

            for auth_handler in app.authentication_strategy.handlers:
                for doc_handler in self._security_schemes_handlers:
                    # We need to activate the service here, if it is a type
                    if not isinstance(auth_handler, AuthenticationHandler):
                        auth_handler = app.services.resolve(auth_handler)

                    for (
                        scheme,
                        security_scheme,
                        req,
                    ) in doc_handler.get_security_schemes(auth_handler):
                        schemes[scheme] = security_scheme

                        if self.security is None:
                            self.security = Security(
                                [], optional=not self.app_requires_auth(app)
                            )
                        self.security.requirements.append(req)

        for name, scheme in schemes.items():
            self.register_security_scheme(name, scheme)  # type: ignore

    def register_security_scheme(self, name: str, scheme: SecurityScheme):
        """
        Adds a security scheme to the OpenAPI documentation components.
        """
        if self.components.security_schemes is None:
            self.components.security_schemes = {}

        self.components.security_schemes[name] = scheme

    def register_schema_for_type(self, object_type: Type) -> Reference:
        stored_ref = self._get_stored_reference(object_type, None)
        if stored_ref:
            return stored_ref

        schema = self.get_schema_by_type(object_type)
        if isinstance(schema, Schema):
            return self._register_schema(schema, self.get_type_name(object_type))
        return schema

    def _register_schema(self, schema: Schema, name: str) -> Reference:
        if self.components.schemas is None:
            self.components.schemas = {}

        if name in self.components.schemas and self.components.schemas[name] != schema:
            counter = 0
            base_name = name
            while name in self.components.schemas:
                counter += 1
                name = f"{base_name}{counter}"

        self.components.schemas[name] = schema
        return Reference(f"#/components/schemas/{name}")

    def _can_handle_class_type(self, object_type: Type) -> bool:
        return (
            any(
                handler.handles_type(object_type)
                for handler in self._object_types_handlers
            )
            or object_type in self._types_schemas
        )

    def _handle_type_having_explicit_schema(self, object_type: Type) -> Reference:
        schema = self._types_schemas[object_type]
        type_name = self.get_type_name(object_type, {})
        reference = self._register_schema(
            schema,
            type_name,
        )
        self._objects_references[object_type] = reference
        self._objects_references[type_name] = reference
        return reference

    def _get_schema_for_class(self, object_type: Type) -> Reference:
        assert self._can_handle_class_type(object_type)

        if object_type in self._types_schemas:
            return self._handle_type_having_explicit_schema(object_type)

        for handler in self._object_types_handlers:
            if not handler.handles_type(object_type):
                continue
            try:
                handler.set_type_schema(object_type, self)
            except NotImplementedError:
                pass
            else:
                return self._handle_type_having_explicit_schema(object_type)

        required: list[str] = []
        properties: dict[str, Schema | Reference] = {}

        for field in self.get_fields(object_type):
            is_optional, child_type = check_union(field.type)
            if not is_optional:
                required.append(field.name)

            if isinstance(child_type, str):
                # this is a forward reference, we need to obtain a full type here
                annotations = get_type_hints(object_type)

                if field.name in annotations:
                    child_type = annotations[field.name]

            if isinstance(child_type, str):
                # This can happen if the user defined a wrong forward reference.
                raise TypeError(f"Could not obtain the actual type for: {child_type}")

            properties[field.name] = self.get_schema_by_type(
                child_type, root_optional=is_optional
            )

        return self._handle_object_type(object_type, properties, required)

    def _handle_object_type(
        self,
        object_type: Type,
        properties: dict[str, Schema | Reference],
        required: list[str],
        context_type_args: dict[Any, Type] | None = None,
    ) -> Reference:
        return self._handle_object_type_schema(
            object_type,
            context_type_args,
            Schema(
                type=ValueType.OBJECT,
                required=required or None,
                properties=properties,
            ),
        )

    def _handle_object_type_schema(
        self, object_type, context_type_args, schema: Schema
    ):
        type_name = self.get_type_name(object_type, context_type_args)
        reference = self._register_schema(schema, type_name)
        self._objects_references[object_type] = reference
        self._objects_references[type_name] = reference
        return reference

    def _get_stored_reference(
        self, object_type: Type[Any], type_args: dict[Any, Type] | None = None
    ) -> Reference | None:
        if object_type in self._objects_references:
            # if object_type is a generic, it can be like
            # Example[~T] while type_args can have the information: {~T: Foo}
            # in such case; check
            return self._objects_references[object_type]

        if type_args:
            type_name = self.get_type_name(object_type, type_args)
            if type_name in self._objects_references:
                return self._objects_references[type_name]

        return None

    def get_schema_by_type(
        self,
        object_type: Type | Schema,
        type_args: dict[Any, Type] | None = None,
        *,
        root_optional: bool | None = None,
    ) -> Schema | Reference:
        if isinstance(object_type, Schema):
            return object_type

        stored_ref = self._get_stored_reference(object_type, type_args)
        if stored_ref:
            return stored_ref

        is_optional, child_type = check_union(object_type)
        schema = self._get_schema_by_type(child_type, type_args)
        if isinstance(schema, Schema):
            if is_optional or root_optional:
                schema.nullable = True
            else:
                schema.nullable = False
        return schema

    def _get_schema_by_type(
        self, object_type: Type[Any], type_args: dict[Any, Type] | None = None
    ) -> Schema | Reference:
        stored_ref = self._get_stored_reference(object_type, type_args)
        if stored_ref:  # pragma: no cover
            return stored_ref

        # # If this is an Annotated type, try to build a schema from its metadata
        # # first. The Annotated alias carries extra information (in __metadata__)
        # # that may describe the actual response/content type (for example
        # # Annotated[Response, PaginatedSet[Cat]]). If we replace the object
        # # with its origin too early we lose that metadata and produce an
        # # empty/default Schema.
        if self._can_handle_class_type(object_type):
            return self._get_schema_for_class(object_type)

        schema = self._try_get_schema_for_simple_type(object_type)
        if schema:
            return schema

        # Dict, OrderedDict, defaultdict are handled first than GenericAlias
        schema = self._try_get_schema_for_mapping(object_type, type_args)
        if schema:
            return schema

        # List, Set, Tuple are handled first than GenericAlias
        schema = self._try_get_schema_for_iterable(object_type, type_args)
        if schema:
            return schema

        if isinstance(object_type, AnnotatedAlias):
            schema = self._try_get_schema_for_annotated(object_type, type_args)
            if schema:
                return schema

        if isinstance(object_type, GenericAlias) or isinstance(object_type, UnionType):
            schema = self._try_get_schema_for_generic(object_type, type_args)
            if schema:
                return schema

        if inspect.isclass(object_type):
            schema = self._try_get_schema_for_enum(object_type)
        return schema or Schema()

    def _try_get_schema_for_simple_type(self, object_type: Type) -> Schema | None:
        if object_type is str:
            return Schema(type=ValueType.STRING)

        if object_type is bytes:
            return Schema(type=ValueType.STRING, format=ValueFormat.BINARY)

        if object_type is int:
            return Schema(type=ValueType.INTEGER, format=ValueFormat.INT64)

        if object_type is float:
            return Schema(type=ValueType.NUMBER, format=ValueFormat.FLOAT)

        if object_type is bool:
            return Schema(type=ValueType.BOOLEAN)

        if object_type is UUID:
            return Schema(type=ValueType.STRING, format=ValueFormat.UUID)

        if object_type is date:
            return Schema(type=ValueType.STRING, format=ValueFormat.DATE)

        if object_type is datetime:
            return Schema(type=ValueType.STRING, format=ValueFormat.DATETIME)

        return None

    def _try_get_schema_for_iterable(
        self, object_type: Type, context_type_args: dict[Any, Type] | None = None
    ) -> Schema | None:
        if object_type in {list, set, tuple}:
            # the user didn't specify the item type
            return Schema(type=ValueType.ARRAY, items=Schema(type=ValueType.STRING))

        origin = get_origin(object_type)

        if not origin or origin not in {list, set, tuple, collections_abc.Sequence}:
            return None

        # can be List, list[str] or list[str] (Python 3.9),
        # note: it could also be union if it wasn't handled above for dataclasses
        try:
            type_args = typing.get_args(object_type)  # type: ignore
        except AttributeError:  # pragma: no cover
            item_type = str
        else:
            item_type = next(iter(type_args), str)

        if context_type_args and item_type in context_type_args:
            item_type = context_type_args.get(item_type)

        return Schema(
            type=ValueType.ARRAY,
            items=self.get_schema_by_type(item_type, context_type_args),
        )

    def _try_get_schema_for_mapping(
        self, object_type: Type, context_type_args: dict[Any, Type] | None = None
    ) -> Schema | None:
        if object_type in {dict, defaultdict, OrderedDict}:
            # the user didn't specify the key and value types
            return Schema(
                type=ValueType.OBJECT,
                additional_properties=Schema(
                    type=ValueType.STRING,
                ),
            )

        origin = get_origin(object_type)

        if not origin or origin not in {
            dict,
            Dict,
            collections_abc.Mapping,
        }:
            return None

        # can be Dict, dict[str, str] or dict[str, str] (Python 3.9),
        # note: it could also be union if it wasn't handled above for dataclasses
        try:
            _, value_type = typing.get_args(object_type)  # type: ignore
        except AttributeError:  # pragma: no cover
            value_type = str

        if context_type_args and value_type in context_type_args:
            value_type = context_type_args.get(value_type, value_type)

        return Schema(
            type=ValueType.OBJECT,
            additional_properties=self.get_schema_by_type(
                value_type, context_type_args
            ),
        )

    def get_fields(self, object_type: Any) -> list[FieldInfo]:
        for handler in self._object_types_handlers:
            if handler.handles_type(object_type):
                return handler.get_type_fields(
                    object_type, self.register_schema_for_type
                )

        return []

    def _try_get_schema_for_annotated(
        self, object_type: Type, context_type_args: dict[Any, Type] | None = None
    ) -> Schema | Reference | None:
        annotations = [child for child in getattr(object_type, "__metadata__", [])]
        if len(annotations) == 1:
            return self.get_schema_by_type(annotations[0], context_type_args)
        assert (
            None not in annotations
        ), "None is not a valid type for an annotated type with multiple annotations"
        schema = Schema(
            ValueType.OBJECT,
            any_of=[
                self.get_schema_by_type(annotation, context_type_args)
                for annotation in annotations
            ],
        )
        return self._handle_object_type_schema(object_type, context_type_args, schema)

    def _try_get_schema_for_generic(
        self, object_type: Type, context_type_args: dict[Any, Type] | None = None
    ) -> Reference | None:
        origin = get_origin(object_type)

        if origin is dict:
            schema = Schema(type=ValueType.OBJECT)
            return self._handle_object_type_schema(
                object_type, context_type_args, schema
            )

        if origin is Union or origin is UnionType:
            schema = Schema(
                ValueType.OBJECT,
                any_of=[
                    self.get_schema_by_type(child_type, context_type_args)
                    for child_type in typing.get_args(object_type)
                ],
            )
            return self._handle_object_type_schema(
                object_type, context_type_args, schema
            )

        required: list[str] = []
        properties: dict[str, Schema | Reference] = {}

        args = typing.get_args(object_type)
        parameters = origin.__parameters__
        type_args = dict(zip(parameters, args))

        for field in self.get_fields(origin):
            is_optional, child_type = check_union(field.type)
            if not is_optional:
                required.append(field.name)

            if isinstance(child_type, str):
                warnings.warn(
                    f"The return type {object_type!r} contains a forward reference "
                    f"for {field.name}. Forward references in Generic types are not "
                    "supported for automatic generation of OpenAPI Documentation. "
                    "For more information see "
                    "https://www.neoteroi.dev/blacksheep/openapi/",
                    UserWarning,
                )
                return None

            if child_type in type_args:
                # example:
                # class Foo(Generic[T]):
                #    item: T
                child_type = type_args.get(child_type)

            properties[field.name] = self.get_schema_by_type(child_type, type_args)

        return self._handle_object_type(
            object_type, properties, required, context_type_args
        )

    def _try_get_schema_for_enum(self, object_type: Type) -> Schema | None:
        if issubclass(object_type, IntEnum):
            return Schema(type=ValueType.INTEGER, enum=[v.value for v in object_type])
        if issubclass(object_type, Enum):
            return Schema(type=ValueType.STRING, enum=[v.value for v in object_type])
        return None

    def _get_body_binder(self, handler: Any) -> BodyBinder | None:
        return next(
            (binder for binder in handler.binders if isinstance(binder, BodyBinder)),
            None,
        )

    def _get_binder_by_name(self, handler: Any, name: str) -> Binder | None:
        return next(
            (binder for binder in handler.binders if binder.parameter_name == name),
            None,
        )

    def _get_body_binder_content_type(
        self,
        body_binder: BodyBinder,
        body_examples: dict[str, Example | Reference] | None,
    ) -> dict[str, MediaType]:
        return {
            possible_type: MediaType(
                schema=self.get_schema_by_type(body_binder.expected_type),
                examples=body_examples,
            )
            for possible_type in body_binder.content_type.split(";")
        }

    def get_request_body(self, handler: Any) -> RequestBody | Reference | None:
        if not hasattr(handler, "binders"):
            return None
        # TODO: improve this code to support more scenarios!
        # https://github.com/Neoteroi/BlackSheep/issues/546
        body_binder = self._get_body_binder(handler)

        if body_binder is None:
            return None

        docs = self.get_handler_docs(handler)
        body_info = docs.request_body if docs else None

        body_examples: dict[str, Example | Reference] | None = (
            {key: Example(value=value) for key, value in body_info.examples.items()}
            if body_info and body_info.examples
            else None
        )

        return RequestBody(
            content=self._get_body_binder_content_type(body_binder, body_examples),
            required=body_binder.required,
            description=body_info.description if body_info else None,
        )

    def get_parameter_location_for_binder(
        self, binder: Binder
    ) -> ParameterLocation | None:
        if isinstance(binder, RouteBinder):
            return ParameterLocation.PATH
        if isinstance(binder, QueryBinder):
            return ParameterLocation.QUERY
        if isinstance(binder, CookieBinder):
            return ParameterLocation.COOKIE

        # exclude those specific header value per documentation
        # https://swagger.io/docs/specification/describing-parameters/#header-parameters
        if isinstance(binder, HeaderBinder) and binder.parameter_name.lower() not in [
            "content-type",
            "accept",
            "authorization",
        ]:
            return ParameterLocation.HEADER
        return None

    def _parameter_source_to_openapi_obj(
        self, value: ParameterSource
    ) -> ParameterLocation:
        return ParameterLocation[value.value.upper()]

    def get_parameters(self, handler: Any) -> list[Parameter | Reference] | None:
        if not hasattr(handler, "binders"):
            return None
        binders: list[Binder] = handler.binders
        parameters: dict[str, Parameter | Reference] = {}

        docs = self.get_handler_docs(handler)
        parameters_info = (docs.parameters if docs else None) or dict()

        for binder in binders:
            if binder.__class__ in self._binder_docs:
                self._handle_binder_docs(binder, parameters)
                continue

            location = self.get_parameter_location_for_binder(binder)

            if not location:
                # the binder is used for something that is not a parameter
                # expressed in OpenAPI Docs (e.g. a DI service)
                continue

            if location == ParameterLocation.PATH:
                required = True
            else:
                required = binder.required and binder.default is empty

            # did the user specified information about the parameter?
            param_info = parameters_info.get(binder.parameter_name)

            parameters[binder.parameter_name] = Parameter(
                name=binder.parameter_name,
                in_=location,
                required=required or None,
                schema=self.get_schema_by_type(binder.expected_type),
                description=param_info.description if param_info else "",
                example=param_info.example if param_info else None,
            )

        for key, param_info in parameters_info.items():
            if key not in parameters:
                parameters[key] = Parameter(
                    name=key,
                    in_=self._parameter_source_to_openapi_obj(
                        param_info.source or ParameterSource.QUERY
                    ),
                    required=param_info.required,
                    schema=(
                        self.get_schema_by_type(param_info.value_type)
                        if param_info.value_type
                        else None
                    ),
                    description=param_info.description,
                    example=param_info.example,
                )

        return list(parameters.values())

    def get_handler_security(self, handler: Any) -> list[SecurityRequirement | None]:
        docs = self.get_handler_docs(handler)
        if docs and docs.security:
            return [SecurityRequirement(s.name, s.value) for s in docs.security]

        # is the handler decorated with the built-in @auth decorator?
        auth_schemes = getattr(handler, "auth_schemes", None)
        if auth_schemes:
            # TODO: how to support scope names in the list of security requirements?
            return [
                SecurityRequirement(auth_scheme, []) for auth_scheme in auth_schemes
            ]

        # is the handler decorated with "allow_anonymous"?
        if getattr(handler, "allow_anonymous", False):
            # return an empty array to specify the endpoint does not require
            # authentication
            return []

        return None

    def _get_media_type_from_content_doc(self, content_doc: ContentInfo) -> MediaType:
        media_type = MediaType()

        if content_doc.type:
            media_type.schema = self.get_schema_by_type(content_doc.type)

        examples = content_doc.examples
        if examples:
            if len(examples) == 1:
                example_item = examples[0]
                if isinstance(example_item, ResponseExample):
                    media_type.example = self.normalize_example(example_item.value)
                else:
                    media_type.example = self.normalize_example(example_item)
            else:
                examples_doc: dict[str, Example | Reference] = {}

                for index, example in enumerate(examples):
                    # support something that is not a ResponseExample,
                    # to offer a more concise API
                    if not isinstance(example, ResponseExample):
                        example = ResponseExample(example)

                    if not example.name:
                        example.name = f"example {index}"
                    examples_doc[example.name] = Example(
                        summary=example.summary,
                        description=example.description,
                        value=self.normalize_example(example.value),
                    )
                media_type.examples = examples_doc

        return media_type

    def _get_content_from_response_info(
        self, response_content: list[ContentInfo | None]
    ) -> dict[str, MediaType | Reference] | None:
        if not response_content:
            return None

        oad_content: dict[str, MediaType | Reference] = {}

        for content in response_content:
            if content.content_type not in oad_content:
                oad_content[content.content_type] = (
                    self._get_media_type_from_content_doc(content)
                )
            else:
                raise DuplicatedContentTypeDocsException(content.content_type)

        return oad_content

    def _get_headers_from_response_info(
        self, headers: dict[str, HeaderInfo | None]
    ) -> dict[str, Header | Reference] | None:
        if headers is None:
            return None

        return {
            key: Header(value.description, self.get_schema_by_type(value.type))
            for key, value in headers.items()
        }

    def get_responses(self, handler: Any) -> dict[str, ResponseDoc | None]:
        docs = self.get_handler_docs(handler)
        data = docs.responses if docs else None

        # common responses used by the whole application, like responses sent in case
        # of error
        responses = {
            response_status_to_str(key): value
            for key, value in self.common_responses.items()
        }

        if not data:
            # try to generate automatically from the handler's return type annotations
            return_type = getattr(handler, "return_type", ...)

            if return_type is not ...:
                # automatically set response content for status 200,
                # if the user wants major control, it's necessary to use the decorators
                # responses[response_status_to_str(200)] = return_type
                if data is None:
                    data = {}

                if (
                    isinstance(return_type, AnnotatedAlias)
                    and return_type.__metadata__[0] is None
                ):
                    return_type = None

                if return_type is None:
                    # the user explicitly marked the request handler as returning None,
                    # document therefore HTTP 204 No Content
                    data["204"] = ResponseInfo("Success response", content=[])
                else:
                    # if the return type is optional, generate automatically
                    # the "404" response
                    is_optional, child_type = check_union(return_type)

                    data["200"] = ResponseInfo(
                        "Success response",
                        content=[ContentInfo(child_type, examples=[])],
                    )

                    if is_optional and self.handle_optional_response_with_404:
                        data["404"] = ResponseInfo("Object not found")
            else:
                return responses

        responses.update(
            {
                response_status_to_str(key): (
                    ResponseDoc(
                        description=value.description,
                        content=self._get_content_from_response_info(value.content),
                        headers=self._get_headers_from_response_info(value.headers),
                    )
                    if isinstance(value, ResponseInfo)
                    else ResponseDoc(description=value)
                )
                for key, value in data.items()
            }
        )
        return responses

    def on_docs_generated(self, docs: OpenAPI) -> None:
        docs.servers = self.servers

        self.events.on_docs_created.fire_sync(docs)

    def _merge_documentation(
        self,
        handler,
        endpoint_docs: EndpointDocs,
        docstring_info: DocstringInfo,
    ) -> None:
        if not endpoint_docs.description and docstring_info.description:
            endpoint_docs.description = docstring_info.description

        if not endpoint_docs.summary and docstring_info.summary:
            endpoint_docs.summary = docstring_info.summary

        for param_name, param_info in docstring_info.parameters.items():
            if endpoint_docs.parameters is None:
                endpoint_docs.parameters = {}

            # did the user specify parameter information explicitly, using @docs?
            matching_parameter = endpoint_docs.parameters.get(param_name)

            if matching_parameter is None:
                matching_binder = self._get_binder_by_name(handler, param_name)

                if isinstance(matching_binder, BodyBinder):
                    if endpoint_docs.request_body is None:
                        endpoint_docs.request_body = RequestBodyInfo(
                            description=param_info.description
                        )
                    elif not endpoint_docs.request_body.description:
                        endpoint_docs.request_body.description = param_info.description
                    continue

                if is_ignored_parameter(param_name, matching_binder):
                    # this must not be documented in OpenAPI Documentation!
                    continue

                assert isinstance(endpoint_docs.parameters, dict)
                endpoint_docs.parameters[param_name] = ParameterInfo(
                    value_type=param_info.value_type,
                    required=param_info.required,
                    description=param_info.description,
                    source=param_info.source or ParameterSource.QUERY,
                )
            else:
                matching_parameter.description = param_info.description

                if (
                    matching_parameter.value_type is None
                    and param_info.value_type is not None
                ):
                    matching_parameter.value_type = param_info.value_type

    def _apply_docstring(self, handler, docs: EndpointDocs | None) -> None:
        if not self.use_docstrings:  # pragma: no cover
            return
        docstring_info = get_handler_docstring_info(handler)

        if docstring_info is not None:
            if docs is None:
                docs = self.get_handler_docs_or_set(handler)
            self._merge_documentation(handler, docs, docstring_info)

    def get_operation_id(self, docs: EndpointDocs | None, handler) -> str:
        return handler.__name__

    def get_routes_docs(
        self, router: Router, path_prefix: str = ""
    ) -> dict[str, PathItem]:
        """Obtains a documentation object from the routes defined in a router."""
        paths_doc: dict[str, PathItem] = {}
        raw_dict = self.router_to_paths_dict(router, lambda route: route)

        for path, conf in raw_dict.items():
            path_item = PathItem()

            for method, route in conf.items():
                handler = self._get_request_handler(route)
                docs = self.get_handler_docs(handler)
                self._apply_docstring(handler, docs)

                operation = Operation(
                    description=self.get_description(handler),
                    summary=self.get_summary(handler),
                    responses=self.get_responses(handler) or {},
                    operation_id=self.get_operation_id(docs, handler),
                    parameters=self.get_parameters(handler),
                    request_body=self.get_request_body(handler),
                    deprecated=self.is_deprecated(handler),
                    tags=self.get_handler_tags(handler),
                    security=self.get_handler_security(handler),
                )
                if docs and docs.on_created:
                    docs.on_created(self, operation)

                self.events.on_operation_created.fire_sync(operation)
                setattr(path_item, method, operation)

            path_value = path_prefix + path
            if path_value != "/" and path_value.endswith("/"):
                path_value = path_value.rstrip("/")
            paths_doc[path_value] = path_item

        self.events.on_paths_created.fire_sync(paths_doc)
        return paths_doc

    def set_binder_docs(
        self,
        binder_type: Type[Binder],
        params_docs: Iterable[Parameter | Reference],
    ):
        """
        Configures parameters documentation for a given binder type. A binder can
        read values from one or more input parameters, this is why this method supports
        an iterable of Parameter or Reference objects. In most use cases, it is
        desirable to use a Parameter here. Reference objects are configured
        automatically when the documentation is built.
        """
        self._binder_docs[binder_type] = params_docs

    def _handle_binder_docs(
        self, binder: Binder, parameters: dict[str, Parameter | Reference]
    ):
        params_docs = self._binder_docs[binder.__class__]

        for i, param_doc in enumerate(params_docs):
            parameters[f"{binder.__class__.__qualname__}_{i}"] = param_doc

    def _optimize_binders_docs(self):
        """
        Optimizes the documentation for custom binders to use references and
        components.parameters, instead of duplicating parameters documentation in each
        operation where they are used.
        """
        new_dict = {}
        params_docs: Iterable[Parameter | Reference]

        for key, params_docs in self._binder_docs.items():
            new_docs: list[Reference] = []

            for param in params_docs:
                if isinstance(param, Reference):
                    new_docs.append(param)
                else:
                    if self.components.parameters is None:
                        self.components.parameters = {}

                    self.components.parameters[param.name] = param
                    new_docs.append(Reference(f"#/components/parameters/{param.name}"))

            new_dict[key] = new_docs

        self._binder_docs = new_dict



"""
This module provides classes and functions to enable distributed tracing and logging
using OpenTelemetry.

Additional dependencies:
    pip install opentelemetry-distro
    opentelemetry-bootstrap --action=install

Features:

- An `use_open_telemetry` function that can be used to apply useful configuration.
- OTELMiddleware: Middleware for automatic tracing of HTTP requests.
- Environment-based configuration for OpenTelemetry resource attributes.
- Logging and tracing setup using user-provided exporters.
- Context manager and decorator utilities for tracing custom operations and function
  calls.

Usage:
    from blacksheep.server.otel import use_open_telemetry

    # Configure log_exporter and span_exporter as needed
    use_open_telemetry(app, log_exporter, span_exporter)
"""

import logging
import os
from contextlib import contextmanager
from functools import wraps
from typing import Awaitable, Callable

from opentelemetry import trace
from opentelemetry.instrumentation.logging import LoggingInstrumentor
from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor, LogExporter
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, SpanExporter
from opentelemetry.trace import SpanKind

from blacksheep import Application
from blacksheep.messages import Request, Response
from blacksheep.middlewares import MiddlewareCategory
from blacksheep.server.env import get_env

ExceptionHandler = Callable[[Request, Exception], Awaitable[Response]]


class OTELMiddleware:
    """
    Middleware configuring OpenTelemetry for all web requests.
    """

    def __init__(self, exc_handler: ExceptionHandler) -> None:
        self._exc_handler = exc_handler
        self._tracer = trace.get_tracer(__name__)

    async def __call__(self, request: Request, handler):
        path = request.url.path.decode("utf8")
        method = request.method
        with self._tracer.start_as_current_span(
            f"{method} {path}", kind=SpanKind.SERVER
        ) as span:
            try:
                response = await handler(request)
            except Exception as exc:
                # This approach is correct because it supports controlling the response
                # using exceptions. Unhandled exceptions are handled by the Span.
                response = await self._exc_handler(request, exc)

            self.set_span_attributes(span, request, response, path)
            return response

    def set_span_attributes(
        self, span: trace.Span, request: Request, response: Response, path: str
    ) -> None:
        """
        Configure the attributes on the span for a given request-response cycle.
        """
        # To reduce cardinality, update the span name to use the
        # route that matched the request
        route = request.route  # type: ignore
        span.update_name(f"{request.method} {route}")

        span.set_attribute("http.status_code", response.status)
        span.set_attribute("http.method", request.method)
        span.set_attribute("http.path", path)
        span.set_attribute("http.url", request.url.value.decode())
        span.set_attribute("http.route", route)
        span.set_attribute("http.status_code", response.status)
        span.set_attribute("client.ip", request.original_client_ip)

        if response.status >= 400:
            span.set_status(trace.Status(trace.StatusCode.ERROR))


def _configure_logging(log_exporter: LogExporter, span_exporter: SpanExporter):
    """
    - Set up a custom LoggerProvider and attach a BatchLogRecordProcessor with the
      provided log_exporter.
    - Set the log level for the "opentelemetry" logger to WARNING to reduce noise.
    - Add a LoggingHandler to the root logger, ensuring OpenTelemetry logs are
      processed
    - Instrument logging with LoggingInstrumentor().instrument(set_logging_format=True)
      to ensure logs are formatted and correlated with traces.
    - Set up the tracer provider and attaches a BatchSpanProcessor for the given
      span_exporter.
    """
    log_provider = LoggerProvider()
    log_provider.add_log_record_processor(BatchLogRecordProcessor(log_exporter))
    logging.getLogger("opentelemetry").setLevel(logging.WARNING)
    logging.getLogger().addHandler(
        LoggingHandler(level=logging.NOTSET, logger_provider=log_provider)
    )

    LoggingInstrumentor().instrument(set_logging_format=True)

    trace.set_tracer_provider(TracerProvider())
    trace.get_tracer_provider().add_span_processor(
        BatchSpanProcessor(span_exporter)
    )  # type: ignore


def set_attributes(
    service_name: str,
    service_namespace: str = "default",
    env: str = "",
):
    """
    Sets the OTEL_RESOURCE_ATTRIBUTES environment variable with service metadata
    for OpenTelemetry.

    Args:
        service_name (str): The name of the service.
        service_namespace (str, optional): The namespace of the service. Defaults to
                                           "default".
        env (str, optional): The deployment environment. If not provided, it is
                            determined from the environment.

    Returns:
        None
    """
    if not env:
        env = get_env()
    os.environ["OTEL_RESOURCE_ATTRIBUTES"] = (
        f"service.name={service_name},"
        f"service.namespace={service_namespace},"
        f"deployment.environment={env}"
    )


def use_open_telemetry(
    app: Application,
    log_exporter: LogExporter,
    span_exporter: SpanExporter,
    middleware: OTELMiddleware | None = None,
):
    """
    Configures OpenTelemetry tracing and logging for a BlackSheep application.

    This function sets up OpenTelemetry log and span exporters, configures resource
    attributes, and injects OTEL middleware for automatic tracing of HTTP requests.
    It also patches the router to track matched route patterns and ensures proper
    shutdown of the tracer provider on application stop.

    Args:
        app (Application): The BlackSheep application instance.
        log_exporter (LogExporter): The OpenTelemetry log exporter to use.
        span_exporter (SpanExporter): The OpenTelemetry span exporter to use.
        middleware (optional OTELMiddleware): Custom OTEL middleware instance.
            If not provided, the default OTELMiddleware is used.

    Returns:
        None
    """
    if os.getenv("OTEL_RESOURCE_ATTRIBUTES") is None:
        # set a default value
        set_attributes("blacksheep-app")

    _configure_logging(log_exporter, span_exporter)

    # Insert the middleware at the beginning of the middlewares list
    app.middlewares.append(
        middleware or OTELMiddleware(app.handle_request_handler_exception),
        MiddlewareCategory.INIT,
        -10,
    )

    @app.on_start
    async def on_start(app):
        # Patch the router to keep track of the route pattern that matched the request,
        # if any.
        # https://www.neoteroi.dev/blacksheep/routing/#how-to-track-routes-that-matched-a-request
        def wrap_get_route_match(fn):
            @wraps(fn)
            def get_route_match(request):
                match = fn(request)
                request.route = match.pattern.decode() if match else "Not Found"
                return match

            return get_route_match

        app.router.get_match = wrap_get_route_match(app.router.get_match)

    @app.on_stop
    async def on_stop(app):
        # Try calling shutdown() on app stop to flush all remaining spans.
        try:
            trace.get_tracer_provider().shutdown()
        except AttributeError:
            pass


@contextmanager
def client_span_context(
    operation_name: str, attributes: dict[str, str], *args, **kwargs
):
    tracer = trace.get_tracer(__name__)
    with tracer.start_as_current_span(operation_name, kind=SpanKind.CLIENT) as span:
        span.set_attributes(attributes)
        for i, value in enumerate(args):
            span.set_attribute(f"@arg{i}", str(value))
        for key, value in kwargs.items():
            span.set_attribute(f"@{key}", str(value))
        try:
            yield
        except Exception as ex:
            span.record_exception(ex)
            span.set_attribute("ERROR", str(ex))
            span.set_attribute("http.status_code", 500)
            span.set_status(trace.Status(trace.StatusCode.ERROR))
            raise


def logcall(component="Service"):
    """
    Wraps a function to log each call using OpenTelemetry, as SpanKind.CLIENT.
    """

    def log_decorator(fn):
        @wraps(fn)
        async def wrapper(*args, **kwargs):
            with client_span_context(
                fn.__name__, {"component": component}, *args, **kwargs
            ):
                return await fn(*args, **kwargs)

        return wrapper

    return log_decorator



import pytest

from blacksheep.messages import Response
from blacksheep.middlewares import (
    CategorizedMiddleware,
    MiddlewareCategory,
    MiddlewareList,
)


# Test fixtures
async def middleware_a() -> Response: ...


async def middleware_b() -> Response: ...


async def middleware_c() -> Response: ...


async def middleware_d() -> Response: ...


async def middleware_e() -> Response: ...


class TestMiddlewareList:
    def test_init(self):
        """Test MiddlewareList initialization"""
        ml = MiddlewareList()
        assert len(ml) == 0
        assert not ml
        assert ml._is_sorted is True
        assert ml._configured is False

    def test_append_default_category(self):
        """Test appending middleware with default category"""
        ml = MiddlewareList()
        ml.append(middleware_a)

        assert len(ml) == 1
        assert ml[0] is middleware_a

        # Check internal structure
        categorized = list(ml.items())[0]
        assert categorized.middleware is middleware_a
        assert categorized.category == MiddlewareCategory.BUSINESS
        assert categorized.priority == 0

    def test_append_with_category_and_priority(self):
        """Test appending middleware with specific category and priority"""
        ml = MiddlewareList()
        ml.append(middleware_a, MiddlewareCategory.AUTH, priority=5)

        categorized = list(ml.items())[0]
        assert categorized.middleware is middleware_a
        assert categorized.category == MiddlewareCategory.AUTH
        assert categorized.priority == 5

    def test_insert(self):
        """Test inserting middleware (legacy support)"""
        ml = MiddlewareList()
        ml.append(middleware_a)
        ml.insert(0, middleware_b)

        assert len(ml) == 2
        # insert uses INIT category with priority -1, so it should come first
        assert ml[0] is middleware_b
        assert ml[1] is middleware_a

    def test_extend(self):
        """Test extending with multiple middlewares"""
        ml = MiddlewareList()
        ml.extend([middleware_a, middleware_b, middleware_c])

        assert len(ml) == 3
        assert ml[0] is middleware_a
        assert ml[1] is middleware_b
        assert ml[2] is middleware_c

    def test_clear(self):
        """Test clearing all middlewares"""
        ml = MiddlewareList()
        ml.append(middleware_a)
        ml.append(middleware_b)

        assert len(ml) == 2
        ml.clear()
        assert len(ml) == 0
        assert not ml
        assert ml._is_sorted is True

    def test_category_sorting(self):
        """Test that middlewares are sorted by category"""
        ml = MiddlewareList()

        # Add middlewares in reverse category order
        ml.append(middleware_a, MiddlewareCategory.MESSAGE)  # 60
        ml.append(middleware_b, MiddlewareCategory.AUTH)  # 30
        ml.append(middleware_c, MiddlewareCategory.INIT)  # 10
        ml.append(middleware_d, MiddlewareCategory.BUSINESS)  # 50

        # Should be sorted by category value
        middlewares = ml.to_list()
        assert middlewares[0] is middleware_c  # INIT (10)
        assert middlewares[1] is middleware_b  # AUTH (30)
        assert middlewares[2] is middleware_d  # BUSINESS (50)
        assert middlewares[3] is middleware_a  # RESPONSE (60)

    def test_priority_sorting_within_category(self):
        """Test that middlewares are sorted by priority within same category"""
        ml = MiddlewareList()

        # Add middlewares in same category with different priorities
        ml.append(middleware_a, MiddlewareCategory.AUTH, priority=3)
        ml.append(middleware_b, MiddlewareCategory.AUTH, priority=1)
        ml.append(middleware_c, MiddlewareCategory.AUTH, priority=2)

        middlewares = ml.to_list()
        assert middlewares[0] is middleware_b  # priority 1
        assert middlewares[1] is middleware_c  # priority 2
        assert middlewares[2] is middleware_a  # priority 3

    def test_mixed_category_and_priority_sorting(self):
        """Test sorting with both different categories and priorities"""
        ml = MiddlewareList()

        ml.append(middleware_a, MiddlewareCategory.BUSINESS, priority=2)
        ml.append(middleware_b, MiddlewareCategory.AUTH, priority=3)
        ml.append(middleware_c, MiddlewareCategory.AUTH, priority=1)
        ml.append(middleware_d, MiddlewareCategory.INIT, priority=5)

        middlewares = ml.to_list()
        assert middlewares[0] is middleware_d  # INIT (10), priority 5
        assert middlewares[1] is middleware_c  # AUTH (30), priority 1
        assert middlewares[2] is middleware_b  # AUTH (30), priority 3
        assert middlewares[3] is middleware_a  # BUSINESS (50), priority 2

    def test_lazy_sorting(self):
        """Test that sorting only happens when needed"""
        ml = MiddlewareList()
        ml.append(middleware_a, MiddlewareCategory.MESSAGE)

        # After append, should not be sorted
        assert ml._is_sorted is False

        # Accessing should trigger sort
        _ = ml[0]
        assert ml._is_sorted is True

        # Adding again should mark as unsorted
        ml.append(middleware_b, MiddlewareCategory.INIT)
        assert ml._is_sorted is False

    def test_configuration_lock(self):
        """Test that middlewares cannot be added after configuration"""
        ml = MiddlewareList()
        ml.append(middleware_a)

        # Mark as configured
        ml._mark_configured()

        # Should raise errors for modification attempts
        with pytest.raises(
            RuntimeError, match="Cannot add middlewares after configuration is complete"
        ):
            ml.append(middleware_b)

        with pytest.raises(
            RuntimeError, match="Cannot add middlewares after configuration is complete"
        ):
            ml.insert(0, middleware_b)

        with pytest.raises(
            RuntimeError, match="Cannot add middlewares after configuration is complete"
        ):
            ml.extend([middleware_b])

    def test_iterator_behavior(self):
        """Test iterator returns middleware functions in sorted order"""
        ml = MiddlewareList()
        ml.append(middleware_a, MiddlewareCategory.MESSAGE)
        ml.append(middleware_b, MiddlewareCategory.INIT)

        middlewares = list(ml)
        assert middlewares[0] is middleware_b  # INIT comes first
        assert middlewares[1] is middleware_a  # RESPONSE comes second

    def test_bool_behavior(self):
        """Test boolean evaluation of MiddlewareList"""
        ml = MiddlewareList()
        assert not ml  # Empty list is False

        ml.append(middleware_a)
        assert ml  # Non-empty list is True

    def test_getitem_behavior(self):
        """Test indexing behavior"""
        ml = MiddlewareList()
        ml.append(middleware_a, MiddlewareCategory.BUSINESS)
        ml.append(middleware_b, MiddlewareCategory.INIT)

        # Should return middlewares in sorted order
        assert ml[0] is middleware_b  # INIT comes first
        assert ml[1] is middleware_a  # BUSINESS comes second

    def test_items_method(self):
        """Test items() method returns CategorizedMiddleware objects"""
        ml = MiddlewareList()
        ml.append(middleware_a, MiddlewareCategory.AUTH, priority=5)

        items = list(ml.items())
        assert len(items) == 1
        assert isinstance(items[0], CategorizedMiddleware)
        assert items[0].middleware is middleware_a
        assert items[0].category == MiddlewareCategory.AUTH
        assert items[0].priority == 5

    def test_to_list_method(self):
        """Test to_list() method returns sorted middleware functions"""
        ml = MiddlewareList()
        ml.append(middleware_a, MiddlewareCategory.MESSAGE)
        ml.append(middleware_b, MiddlewareCategory.INIT)

        middlewares = ml.to_list()
        assert isinstance(middlewares, list)
        assert middlewares[0] is middleware_b  # INIT first
        assert middlewares[1] is middleware_a  # RESPONSE second

    def test_mark_configured_sorts_middlewares(self):
        """Test that _mark_configured() triggers sorting"""
        ml = MiddlewareList()
        ml.append(middleware_a, MiddlewareCategory.MESSAGE)
        ml.append(middleware_b, MiddlewareCategory.INIT)

        assert ml._is_sorted is False
        assert ml._configured is False

        ml._mark_configured()

        assert ml._is_sorted is True
        assert ml._configured is True

    def test_insert_uses_init_category(self):
        """Test that insert() uses INIT category with priority -1"""
        ml = MiddlewareList()
        ml.append(middleware_a, MiddlewareCategory.INIT, priority=0)
        ml.insert(0, middleware_b)  # Should get INIT category, priority -1

        # middleware_b should come first due to lower priority (-1 vs 0)
        assert ml[0] is middleware_b
        assert ml[1] is middleware_a

        # Check the categorized middleware
        items = list(ml.items())
        inserted_middleware = next(
            item for item in items if item.middleware is middleware_b
        )
        assert inserted_middleware.category == MiddlewareCategory.INIT
        assert inserted_middleware.priority == -1

    def test_empty_extend(self):
        """Test extending with empty list"""
        ml = MiddlewareList()
        ml.extend([])

        assert len(ml) == 0
        assert not ml



import pytest

from blacksheep.contents import ServerSentEvent
from blacksheep.server.sse import ServerSentEventsResponse


@pytest.mark.asyncio
async def test_server_sent_events_response_streams_events_1():
    # Arrange: create a simple events provider
    async def events_provider():
        yield ServerSentEvent(data={"message": "hello"}, event="greeting", id="1")
        yield ServerSentEvent(data={"message": "world"}, event="greeting", id="2")

    response = ServerSentEventsResponse(events_provider)
    content = response.content
    assert content is not None
    # Collect all bytes from the streamed content
    result = b""
    async for chunk in content.get_parts():  # type: ignore
        result += chunk
    # Check the output contains the expected SSE format
    assert b"id: 1\n" in result
    assert b"event: greeting\n" in result
    assert b'data: {"message":"hello"}\n' in result
    assert b"id: 2\n" in result
    assert b'data: {"message":"world"}\n' in result
    assert content.type == b"text/event-stream"


@pytest.mark.asyncio
async def test_server_sent_events_response_streams_events_2():
    # Arrange: create a simple events provider
    async def events_provider():
        yield ServerSentEvent(data="hello", event="greeting", id="1")
        yield ServerSentEvent(data="world", event="greeting", id="2")

    response = ServerSentEventsResponse(events_provider)
    content = response.content
    assert content is not None
    # Collect all bytes from the streamed content
    result = b""
    async for chunk in content.get_parts():  # type: ignore
        result += chunk
    # Check the output contains the expected SSE format
    assert b"id: 1\n" in result
    assert b"event: greeting\n" in result
    assert b'data: "hello"\n' in result
    assert b"id: 2\n" in result
    assert b'data: "world"\n' in result
    assert content.type == b"text/event-stream"


@pytest.mark.asyncio
async def test_server_sent_events_response_streams_events_3():
    # Arrange: create a simple events provider
    async def events_provider():
        yield ServerSentEvent(data="hello", event="greeting", id="1")
        yield ServerSentEvent(data="world", event="greeting", id="2")

    response = ServerSentEventsResponse(events_provider)
    content = response.content
    assert content is not None
    # Collect all bytes from the streamed content
    result = b""
    async for chunk in content.get_parts():  # type: ignore
        result += chunk
    # Check the output contains the expected SSE format
    assert b"id: 1\n" in result
    assert b"event: greeting\n" in result
    assert b'data: "hello"\n' in result
    assert b"id: 2\n" in result
    assert b'data: "world"\n' in result
    assert content.type == b"text/event-stream"



import pytest
from essentials.secrets import Secret
from guardpost import Identity

from blacksheep.server.authentication.apikey import (
    APIKey,
    APIKeyAuthentication,
    APIKeyLocation,
    APIKeysProvider,
)
from blacksheep.testing.helpers import get_example_scope
from blacksheep.testing.messages import MockReceive, MockSend
from tests.utils.application import FakeApplication


# Test fixtures
@pytest.fixture
def api_key():
    return APIKey(
        secret=Secret("test-api-key-123", direct_value=True),
        claims={"name": "Test User", "email": "test@example.com"},
        roles=["user", "reader"],
    )


@pytest.fixture
def admin_api_key():
    return APIKey(
        secret=Secret("admin-key-456", direct_value=True),
        claims={"name": "Admin User", "email": "admin@example.com"},
        roles=["admin", "user"],
    )


class MockAPIKeysProvider(APIKeysProvider):
    def __init__(self, keys: list[APIKey]):
        self._keys = keys

    async def get_keys(self) -> list[APIKey]:
        return self._keys


class FailingAPIKeysProvider(APIKeysProvider):
    async def get_keys(self) -> list[APIKey]:
        raise Exception("Database connection failed")


# APIKey tests
def test_api_key_creation():
    api_key = APIKey(
        secret=Secret("test-key", direct_value=True),
        claims={"custom": "claim"},
        roles=["user"],
    )

    assert api_key.claims == {"custom": "claim"}
    assert api_key.roles == ["user"]


def test_api_key_defaults():
    api_key = APIKey(secret=Secret("test-key", direct_value=True))

    assert api_key.claims == {}
    assert api_key.roles == []


def test_api_key_match_success():
    api_key = APIKey(secret=Secret("test-key", direct_value=True))

    assert api_key.match("test-key") is True
    assert api_key.match(Secret("test-key", direct_value=True)) is True


def test_api_key_match_failure():
    api_key = APIKey(secret=Secret("test-key", direct_value=True))

    assert api_key.match("wrong-key") is False
    assert api_key.match(Secret("wrong-key", direct_value=True)) is False


# APIKeyAuthentication creation tests
def test_api_key_authentication_creation_with_keys():
    api_key = APIKey(secret=Secret("test-key", direct_value=True))

    auth = APIKeyAuthentication(api_key, param_name="X-API-Key")
    assert auth.scheme == "APIKey"
    assert auth.param_name == "X-API-Key"
    assert auth.location == APIKeyLocation.HEADER


def test_api_key_authentication_creation_with_provider():
    provider = MockAPIKeysProvider([])
    auth = APIKeyAuthentication(param_name="X-API-Key", keys_provider=provider)
    assert auth._keys_provider is provider


def test_api_key_authentication_creation_with_custom_scheme():
    api_key = APIKey(secret=Secret("test-key", direct_value=True))

    auth = APIKeyAuthentication(api_key, param_name="X-API-Key", scheme="CustomAPI")
    assert auth.scheme == "CustomAPI"


def test_api_key_authentication_creation_with_query_location():
    api_key = APIKey(secret=Secret("test-key", direct_value=True))

    auth = APIKeyAuthentication(api_key, param_name="api_key", location="query")
    assert auth.location == APIKeyLocation.QUERY


def test_api_key_authentication_creation_with_cookie_location():
    api_key = APIKey(secret=Secret("test-key", direct_value=True))

    auth = APIKeyAuthentication(api_key, param_name="api_key", location="cookie")
    assert auth.location == APIKeyLocation.COOKIE


def test_api_key_authentication_creation_requires_keys_or_provider():
    with pytest.raises(
        ValueError, match="Either keys or keys_provider must be provided"
    ):
        APIKeyAuthentication(param_name="X-API-Key")


def test_api_key_authentication_creation_mutual_exclusivity():
    api_key = APIKey(secret=Secret("test-key", direct_value=True))
    provider = MockAPIKeysProvider([])

    with pytest.raises(
        ValueError, match="Cannot specify both static keys and a keys_provider"
    ):
        APIKeyAuthentication(api_key, param_name="X-API-Key", keys_provider=provider)


# Header authentication tests
async def test_api_key_authentication_header_success(app: FakeApplication, api_key):
    app.use_authentication().add(APIKeyAuthentication(api_key, param_name="X-API-Key"))

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"X-API-Key", b"test-api-key-123")],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is True
    assert identity["name"] == "Test User"
    assert identity["email"] == "test@example.com"
    assert identity["roles"] == ["user", "reader"]


async def test_api_key_authentication_header_no_header(app: FakeApplication, api_key):
    app.use_authentication().add(APIKeyAuthentication(api_key, param_name="X-API-Key"))

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is False


async def test_api_key_authentication_header_wrong_key(app: FakeApplication, api_key):
    app.use_authentication().add(APIKeyAuthentication(api_key, param_name="X-API-Key"))

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"X-API-Key", b"wrong-key")],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is False


# Query parameter authentication tests
async def test_api_key_authentication_query_success(app: FakeApplication, api_key):
    app.use_authentication().add(
        APIKeyAuthentication(api_key, param_name="api_key", location="query")
    )

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    await app(
        get_example_scope("GET", "/", query="api_key=test-api-key-123"),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is True
    assert identity["name"] == "Test User"


async def test_api_key_authentication_query_no_param(app: FakeApplication, api_key):
    app.use_authentication().add(
        APIKeyAuthentication(api_key, param_name="api_key", location="query")
    )

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is False


async def test_api_key_authentication_query_wrong_key(app: FakeApplication, api_key):
    app.use_authentication().add(
        APIKeyAuthentication(api_key, param_name="api_key", location="query")
    )

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    await app(
        get_example_scope("GET", "/?api_key=wrong-key"),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is False


# Cookie authentication tests
async def test_api_key_authentication_cookie_success(app: FakeApplication, api_key):
    app.use_authentication().add(
        APIKeyAuthentication(api_key, param_name="api_key", location="cookie")
    )

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Cookie", b"api_key=test-api-key-123")],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is True
    assert identity["name"] == "Test User"


async def test_api_key_authentication_cookie_no_cookie(app: FakeApplication, api_key):
    app.use_authentication().add(
        APIKeyAuthentication(api_key, param_name="api_key", location="cookie")
    )

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is False


async def test_api_key_authentication_cookie_wrong_key(app: FakeApplication, api_key):
    app.use_authentication().add(
        APIKeyAuthentication(api_key, param_name="api_key", location="cookie")
    )

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Cookie", b"api_key=wrong-key")],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is False


# Multiple keys tests
async def test_api_key_authentication_multiple_keys(
    app: FakeApplication, api_key, admin_api_key
):
    app.use_authentication().add(
        APIKeyAuthentication(api_key, admin_api_key, param_name="X-API-Key")
    )

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    # Test with first key
    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"X-API-Key", b"test-api-key-123")],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity["name"] == "Test User"
    assert "admin" not in identity["roles"]

    # Test with second key
    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"X-API-Key", b"admin-key-456")],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity["name"] == "Admin User"
    assert "admin" in identity["roles"]


# Provider tests
async def test_api_key_authentication_with_provider(app: FakeApplication, api_key):
    provider = MockAPIKeysProvider([api_key])
    app.use_authentication().add(
        APIKeyAuthentication(param_name="X-API-Key", keys_provider=provider)
    )

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"X-API-Key", b"test-api-key-123")],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is True
    assert identity["name"] == "Test User"


async def test_api_key_authentication_with_empty_provider(app: FakeApplication):
    provider = MockAPIKeysProvider([])
    app.use_authentication().add(
        APIKeyAuthentication(param_name="X-API-Key", keys_provider=provider)
    )

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"X-API-Key", b"any-key")],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is False


async def test_api_key_authentication_with_failing_provider(app: FakeApplication):
    provider = FailingAPIKeysProvider()
    app.use_authentication().add(
        APIKeyAuthentication(param_name="X-API-Key", keys_provider=provider)
    )

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    # Must not eat exceptions!
    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"X-API-Key", b"any-key")],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 500


# Edge cases and special scenarios
async def test_api_key_authentication_case_sensitive_header(
    app: FakeApplication, api_key
):
    app.use_authentication().add(
        APIKeyAuthentication(api_key, param_name="X-Api-Key")  # Different case
    )

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    # Test with exact case
    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"X-Api-Key", b"test-api-key-123")],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is True


async def test_api_key_authentication_unicode_key(app: FakeApplication):
    unicode_key = APIKey(
        secret=Secret("ключ-тест-123", direct_value=True),  # Cyrillic characters
        claims={"name": "Unicode User"},
    )

    app.use_authentication().add(
        APIKeyAuthentication(unicode_key, param_name="X-API-Key")
    )

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"X-API-Key", "ключ-тест-123".encode("utf-8"))],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is True
    assert identity["name"] == "Unicode User"


async def test_api_key_authentication_empty_key_value(app: FakeApplication, api_key):
    app.use_authentication().add(APIKeyAuthentication(api_key, param_name="X-API-Key"))

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"X-API-Key", b"")],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is False


def test_api_key_authentication_custom_description():
    api_key = APIKey(secret=Secret("test-key", direct_value=True))

    auth = APIKeyAuthentication(
        api_key, param_name="X-API-Key", description="Custom API Key Auth"
    )
    assert auth.description == "Custom API Key Auth"


# Test with APIKeyLocation enum directly
def test_api_key_authentication_with_enum_location():
    api_key = APIKey(secret=Secret("test-key", direct_value=True))

    auth = APIKeyAuthentication(
        api_key, param_name="api_key", location=APIKeyLocation.QUERY
    )
    assert auth.location == APIKeyLocation.QUERY



import os

from blacksheep.utils import truthy


def get_env() -> str:
    return os.environ.get("APP_ENV", "production")


def is_development() -> bool:
    """
    Returns a value indicating whether the application is running for local development.
    This method checks if an `APP_ENV` environment variable is set and its lowercase
    value is either "local", "dev", or "development".
    """
    return get_env().lower() in {"local", "dev", "development"}


def is_production() -> bool:
    """
    Returns a value indicating whether the application is running for the production
    environment (default is true).
    This method checks if an `APP_ENV` environment variable is set and its lowercase
    value is either "prod" or "production".
    """
    return get_env().lower() in {"prod", "production"}


def get_global_route_prefix() -> str:
    """
    Returns the global route prefix, if any, defined by the `APP_ROUTE_PREFIX`
    environment variable.
    """
    return os.environ.get("APP_ROUTE_PREFIX", "")


class EnvironmentSettings:
    _env: str
    _show_error_details: bool
    _mount_auto_events: bool
    _use_default_router: bool
    _add_signal_handler: bool
    _http_scheme: str | None
    _force_https: bool

    def __init__(
        self,
        env: str = "local",
        show_error_details: bool = False,
        mount_auto_events: bool = True,
        use_default_router: bool = True,
        add_signal_handler: bool = False,
        http_scheme: str | None = None,
        force_https: bool = False,
    ) -> None:
        if http_scheme is not None and http_scheme not in {"http", "https"}:
            raise ValueError(
                f"Invalid http_scheme: '{http_scheme}'. Must be 'http' or 'https'."
            )

        self._env = env
        self._show_error_details = show_error_details
        self._mount_auto_events = mount_auto_events
        self._use_default_router = use_default_router
        self._add_signal_handler = add_signal_handler
        self._http_scheme = http_scheme
        self._force_https = force_https

    @classmethod
    def from_env(cls) -> "EnvironmentSettings":
        """
        Creates an EnvironmentSettings instance by reading from environment variables.
        """
        return cls(
            env=get_env(),
            show_error_details=truthy(os.environ.get("APP_SHOW_ERROR_DETAILS", "")),
            mount_auto_events=truthy(os.environ.get("APP_MOUNT_AUTO_EVENTS", "1")),
            use_default_router=truthy(os.environ.get("APP_DEFAULT_ROUTER", "1")),
            add_signal_handler=truthy(os.environ.get("APP_SIGNAL_HANDLER", "")),
            http_scheme=os.environ.get("APP_HTTP_SCHEME"),
            force_https=truthy(os.environ.get("APP_FORCE_HTTPS", "")),
        )

    @property
    def env(self) -> str:
        return self._env

    @property
    def show_error_details(self) -> bool:
        return self._show_error_details

    @property
    def mount_auto_events(self) -> bool:
        return self._mount_auto_events

    @property
    def use_default_router(self) -> bool:
        return self._use_default_router

    @property
    def add_signal_handler(self) -> bool:
        return self._add_signal_handler

    @property
    def http_scheme(self) -> str | None:
        return self._http_scheme

    @property
    def force_https(self) -> bool:
        return self._force_https



from collections.abc import Mapping, MutableSequence


class Header:
    def __init__(self, name: bytes, value: bytes):
        self.name = name
        self.value = value

    def __repr__(self):
        return f"<Header {self.name}: {self.value}>"

    def __iter__(self):
        yield self.name
        yield self.value

    def __eq__(self, other):
        if isinstance(other, Header):
            return other.name.lower() == self.name.lower() and other.value == self.value
        return NotImplemented


class Headers:
    def __init__(self, values: list[tuple[bytes, bytes]] = None):
        if values is None:
            values = []
        self.values = values

    def get(self, name: bytes) -> tuple[bytes, ...]:
        results = []
        name = name.lower()
        for header in self.values:
            if header[0].lower() == name:
                results.append(header[1])
        return tuple(results)

    def get_tuples(self, name: bytes) -> list[tuple[bytes, bytes]]:
        results = []
        name = name.lower()
        for header in self.values:
            if header[0].lower() == name:
                results.append(header)
        return results

    def get_first(self, key: bytes) -> bytes | None:
        key = key.lower()
        for header in self.values:
            if header[0].lower() == key:
                return header[1]
        return None

    def get_single(self, key: bytes) -> bytes:
        results = self.get(key)
        if len(results) > 1:
            raise ValueError("Headers contains more than one header with the given key")
        if len(results) < 1:
            raise ValueError("Headers does not contain one header with the given key")
        return results[0]

    def merge(self, values: list[tuple[bytes, bytes]]):
        for header in values:
            if header is None:
                continue
            self.values.append(header)

    def update(self, values: dict[bytes, bytes]):
        for key, value in values.items():
            self[key] = value

    def items(self):
        yield from self.values

    def clone(self):
        values = []
        for name, value in self.values:
            values.append((name, value))
        return Headers(values)

    def add_many(self, values: dict[bytes, bytes] | list[tuple[bytes, bytes]]):
        if isinstance(values, MutableSequence):
            for item in values:
                self.add(*item)
            return
        if isinstance(values, Mapping):
            for key, value in values.items():
                self.add(key, value)
            return
        raise ValueError("values must be dict[bytes, bytes] or list[Header]")

    @staticmethod
    def _add_to_instance(instance, other):
        if isinstance(other, Headers):
            for value in other:
                instance.add(*value)
            return instance
        if isinstance(other, Header):
            instance.add(other.name, other.value)
            return instance
        if isinstance(other, tuple):
            if len(other) != 2:
                raise ValueError(f"Cannot add, an invalid tuple {str(other)}.")
            instance.add(*other)
            return instance
        if isinstance(other, MutableSequence):
            for value in other:
                if isinstance(value, tuple) and len(value) == 2:
                    instance.add(*value)
                else:
                    raise ValueError(
                        "The sequence contains invalid elements: cannot add "
                        f"{str(value)} to {instance.__class__.__name__}"
                    )
            return instance
        return NotImplemented

    def __add__(self, other):
        return self._add_to_instance(self.clone(), other)

    def __radd__(self, other):
        return self._add_to_instance(self.clone(), other)

    def __iadd__(self, other):
        return self._add_to_instance(self, other)

    def __iter__(self):
        yield from self.values

    def __setitem__(self, key: bytes, value: bytes):
        self.set(key, value)

    def __getitem__(self, item: bytes):
        return self.get(item)

    def keys(self) -> tuple[bytes, ...]:
        results = []
        for name, value in self.values:
            if name not in results:
                results.append(name)
        return tuple(results)

    def add(self, name: bytes, value: bytes):
        self.values.append((name, value))

    def set(self, name: bytes, value: bytes):
        if self.contains(name):
            self.remove(name)
        self.add(name, value)

    def remove(self, key: bytes):
        to_remove = []
        key = key.lower()
        for item in self.values:
            if item[0].lower() == key:
                to_remove.append(item)
        for item in to_remove:
            self.values.remove(item)

    def contains(self, key: bytes) -> bool:
        key = key.lower()
        for name, value in self.values:
            if name.lower() == key:
                return True
        return False

    def __delitem__(self, key: bytes):
        self.remove(key)

    def __contains__(self, key: bytes) -> bool:
        return self.contains(key)

    def __repr__(self):
        return f"<Headers {self.values}>"



from typing import Awaitable, Callable

from blacksheep.messages import Request, Response
from blacksheep.sessions.abc import Session, SessionSerializer, SessionStore

__all__ = [
    "Session",
    "SessionMiddleware",
    "SessionStore",
    "SessionSerializer",
]


class SessionMiddleware:
    """
    Middleware for managing user sessions in a BlackSheep application.

    This middleware loads the session from the provided session store at the beginning
    of the request, attaches it to the request object, and saves the session back to
    the store if it was modified during request processing.

    Args:
        store (SessionStore): The session store used to load and save session data.

    Usage:
        Add this middleware to your application to enable session support.
    """

    def __init__(self, store: SessionStore) -> None:
        self._store = store

    async def __call__(
        self, request: Request, handler: Callable[[Request], Awaitable[Response]]
    ) -> Response:
        session = await self._store.load(request)
        request.session = session
        response = await handler(request)
        if session.modified:
            await self._store.save(request, response, session)
        return response



import copy
import glob
import inspect
import os
from pathlib import Path


def get_parent_file():
    """
    Returns __file__ of the caller's parent module.
    """
    try:
        return inspect.stack()[2][1]
    except IndexError:
        return ""


def import_child_modules(root_path: Path):
    """
    Import automatically all modules defined
    under a certain package path.
    """
    path = str(root_path)
    modules = [
        os.path.basename(f)[:-3]
        for f in glob.glob(path + "/*.py")
        if not os.path.basename(f).startswith("_")
    ]
    stripped_path = os.path.relpath(path).replace("/", ".").replace("\\", ".")
    for module in modules:
        __import__(stripped_path + "." + module)


def clonefunc(func):
    """
    Clone a function, preserving its name, docstring, annotations and kwdefaults.
    """
    new_func = func.__class__(
        func.__code__,
        func.__globals__,
        func.__name__,
        func.__defaults__,
        func.__closure__,
    )
    new_func.__doc__ = func.__doc__
    new_func.__dict__ = copy.deepcopy(func.__dict__)
    new_func.__annotations__ = copy.deepcopy(func.__annotations__)
    if func.__kwdefaults__:
        new_func.__kwdefaults__ = copy.deepcopy(func.__kwdefaults__)
    return new_func


def all_subclasses(cls):
    """
    Return all subclasses of a class, including those defined in other modules.
    """
    subclasses = set()
    for subclass in cls.__subclasses__():
        subclasses.add(subclass)
        subclasses.update(all_subclasses(subclass))
    return subclasses



from typing import TypedDict

class ASGIScopeInterface(TypedDict):
    type: str
    http_version: str
    server: tuple[str, int]
    client: tuple[str, int]
    scheme: str
    method: str
    path: str
    root_path: str
    raw_path: bytes
    query_string: str
    headers: list[tuple[bytes, bytes]]



import os
from unittest.mock import patch

import pytest

from blacksheep.server.env import EnvironmentSettings


def test_env_settings():
    with patch.dict(os.environ, {"APP_SHOW_ERROR_DETAILS": "1"}):
        env = EnvironmentSettings.from_env()
        assert env.show_error_details is True

    with patch.dict(os.environ, {"APP_SHOW_ERROR_DETAILS": "0"}):
        env = EnvironmentSettings.from_env()
        assert env.show_error_details is False

    with patch.dict(os.environ, {"APP_SHOW_ERROR_DETAILS": "true"}):
        env = EnvironmentSettings.from_env()
        assert env.show_error_details is True


def test_env_settings_env_property():
    with patch.dict(os.environ, {"APP_ENV": "development"}):
        env = EnvironmentSettings.from_env()
        assert env.env == "development"

    with patch.dict(os.environ, {"APP_ENV": "production"}):
        env = EnvironmentSettings.from_env()
        assert env.env == "production"


def test_env_settings_default_env():
    with patch.dict(os.environ, {}, clear=True):
        env = EnvironmentSettings.from_env()
        assert env.env == "production"


def test_env_settings_mount_auto_events():
    with patch.dict(os.environ, {"APP_MOUNT_AUTO_EVENTS": "1"}):
        env = EnvironmentSettings.from_env()
        assert env.mount_auto_events is True

    with patch.dict(os.environ, {"APP_MOUNT_AUTO_EVENTS": "0"}):
        env = EnvironmentSettings.from_env()
        assert env.mount_auto_events is False

    with patch.dict(os.environ, {"APP_MOUNT_AUTO_EVENTS": "false"}):
        env = EnvironmentSettings.from_env()
        assert env.mount_auto_events is False


def test_env_settings_mount_auto_events_default():
    with patch.dict(os.environ, {}, clear=True):
        env = EnvironmentSettings.from_env()
        assert env.mount_auto_events is True


def test_env_settings_use_default_router():
    with patch.dict(os.environ, {"APP_DEFAULT_ROUTER": "1"}):
        env = EnvironmentSettings.from_env()
        assert env.use_default_router is True

    with patch.dict(os.environ, {"APP_DEFAULT_ROUTER": "0"}):
        env = EnvironmentSettings.from_env()
        assert env.use_default_router is False


def test_env_settings_use_default_router_default():
    with patch.dict(os.environ, {}, clear=True):
        env = EnvironmentSettings.from_env()
        assert env.use_default_router is True


def test_env_settings_add_signal_handler():
    with patch.dict(os.environ, {"APP_SIGNAL_HANDLER": "1"}):
        env = EnvironmentSettings.from_env()
        assert env.add_signal_handler is True

    with patch.dict(os.environ, {"APP_SIGNAL_HANDLER": "0"}):
        env = EnvironmentSettings.from_env()
        assert env.add_signal_handler is False


def test_env_settings_add_signal_handler_default():
    with patch.dict(os.environ, {}, clear=True):
        env = EnvironmentSettings.from_env()
        assert env.add_signal_handler is False


def test_env_settings_http_scheme():
    with patch.dict(os.environ, {"APP_HTTP_SCHEME": "http"}):
        env = EnvironmentSettings.from_env()
        assert env.http_scheme == "http"

    with patch.dict(os.environ, {"APP_HTTP_SCHEME": "https"}):
        env = EnvironmentSettings.from_env()
        assert env.http_scheme == "https"


def test_env_settings_http_scheme_default():
    with patch.dict(os.environ, {}, clear=True):
        env = EnvironmentSettings.from_env()
        assert env.http_scheme is None


def test_env_settings_force_https():
    with patch.dict(os.environ, {"APP_FORCE_HTTPS": "1"}):
        env = EnvironmentSettings.from_env()
        assert env.force_https is True

    with patch.dict(os.environ, {"APP_FORCE_HTTPS": "0"}):
        env = EnvironmentSettings.from_env()
        assert env.force_https is False

    with patch.dict(os.environ, {"APP_FORCE_HTTPS": "yes"}):
        env = EnvironmentSettings.from_env()
        assert env.force_https is True


def test_env_settings_force_https_default():
    with patch.dict(os.environ, {}, clear=True):
        env = EnvironmentSettings.from_env()
        assert env.force_https is False


def test_env_settings_constructor():
    env = EnvironmentSettings(
        env="test",
        show_error_details=True,
        mount_auto_events=False,
        use_default_router=False,
        add_signal_handler=True,
        http_scheme="https",
        force_https=True,
    )
    assert env.env == "test"
    assert env.show_error_details is True
    assert env.mount_auto_events is False
    assert env.use_default_router is False
    assert env.add_signal_handler is True
    assert env.http_scheme == "https"
    assert env.force_https is True


def test_env_settings_constructor_defaults():
    env = EnvironmentSettings()
    assert env.env == "local"
    assert env.show_error_details is False
    assert env.mount_auto_events is True
    assert env.use_default_router is True
    assert env.add_signal_handler is False
    assert env.http_scheme is None
    assert env.force_https is False


def test_env_settings_invalid_http_scheme():
    with pytest.raises(ValueError, match="Invalid http_scheme"):
        EnvironmentSettings(http_scheme="ftp")

    with pytest.raises(ValueError, match="Invalid http_scheme"):
        EnvironmentSettings(http_scheme="invalid")



import base64

from itsdangerous import (
    BadSignature,
    Serializer,
    SignatureExpired,
    URLSafeTimedSerializer,
)

from blacksheep.cookies import Cookie
from blacksheep.messages import Request, Response
from blacksheep.sessions.abc import Session, SessionSerializer, SessionStore
from blacksheep.sessions.json import JSONSerializer
from blacksheep.sessions.logs import get_logger
from blacksheep.utils import ensure_str


class CookieSessionStore(SessionStore):
    """
    Session store implementation that saves session data in a signed cookie.

    This store serializes session data, signs it using a secret key, and stores it
    in a cookie on the client side. It supports optional session expiration and
    custom serializers and signers.

    Args:
        secret_key (str): Secret key used to sign session cookies.
        session_cookie (str, optional): Name of the session cookie. Defaults to
            "session".
        serializer (SessionSerializer, optional): Serializer for session data. Defaults
            to JSONSerializer.
        signer (Serializer, optional): Serializer used for signing. Defaults to
            URLSafeTimedSerializer.
        session_max_age (int, optional): Maximum age of the session in seconds.

    Raises:
        ValueError: If session_max_age is provided and is less than 1.
    """

    def __init__(
        self,
        secret_key: str,
        *,
        session_cookie: str = "session",
        serializer: SessionSerializer | None = None,
        signer: Serializer | None = None,
        session_max_age: int | None = None,
    ) -> None:
        self._signer = signer or URLSafeTimedSerializer(secret_key)
        self._serializer = serializer or JSONSerializer()
        self._session_cookie = session_cookie
        self._logger = get_logger()
        if session_max_age is not None and session_max_age < 1:
            raise ValueError("session_max_age must be a positive number greater than 0")
        self.session_max_age = session_max_age

    def _try_read_session(self, raw_value: str) -> Session:
        try:
            if self.session_max_age:
                assert isinstance(self._signer, URLSafeTimedSerializer), (
                    "To use a session_max_age, the configured signer must be of "
                    + " URLSafeTimedSerializer type"
                )
                unsigned_value = self._signer.loads(
                    raw_value, max_age=self.session_max_age
                )
            else:
                unsigned_value = self._signer.loads(raw_value)
        except SignatureExpired:
            self._logger.info("The session signature has expired.")
            return Session()
        except BadSignature:
            self._logger.info("The session signature verification failed.")
            return Session()
        return self._serializer.read(base64.b64decode(unsigned_value).decode("utf8"))

    def _write_session(self, session: Session) -> str:
        payload = base64.b64encode(
            self._serializer.write(session).encode("utf8")
        ).decode()
        return ensure_str(self._signer.dumps(payload))  # type: ignore

    def _prepare_cookie(self, value: str) -> Cookie:
        return Cookie(self._session_cookie, value, path="/", http_only=True)

    async def load(self, request: Request) -> Session:
        current_session_value = request.cookies.get(self._session_cookie, None)
        if current_session_value:
            return self._try_read_session(current_session_value)
        return Session()

    async def save(
        self, request: Request, response: Response, session: Session
    ) -> None:
        response.set_cookie(self._prepare_cookie(self._write_session(session)))



import os
import secrets
import warnings
from typing import Iterable, Sequence

from essentials.secrets import Secret
from itsdangerous import Serializer
from itsdangerous.url_safe import URLSafeSerializer

from blacksheep.baseapp import get_logger

logger = get_logger()


def generate_secret(length: int = 48) -> str:
    return secrets.token_urlsafe(length)


def get_keys() -> list[str]:
    """
    Retrieve or generate application secret keys for data protection.

    This function first attempts to load secret keys from environment variables.
    By default, it looks for variables prefixed with 'APP_SECRET' (e.g., APP_SECRET_1,
    APP_SECRET_2). The prefix can be customized using the BLACKSHEEP_SECRET_PREFIX
    environment variable.

    If no environment variables are found, the function generates three new random
    secrets on the fly. Note that generated secrets are not persisted and will change
    on application restart.

    Returns:
        list[str]: A list of secret keys as plain text strings. Returns keys from
                   environment variables if available, otherwise returns three newly
                   generated secrets.

    Note:
        For production use, always configure secrets via environment variables to
        ensure tokens remain valid across application restarts and when using
        multiple instances.

    Example:
        # Set secrets via environment variables
        export APP_SECRET_1="your-secret-key-1"
        export APP_SECRET_2="your-secret-key-2"

        # Or use a custom prefix
        export BLACKSHEEP_SECRET_PREFIX="MY_SECRET"
        export MY_SECRET_1="your-secret-key-1"
    """
    # if there are environmental variables with keys, use them;
    # by default this kind of env variables would be used:
    # APP_SECRET_1="***"
    # APP_SECRET_2="***"
    # APP_SECRET_3="***"
    app_secrets = []
    env_var_key_prefix = os.environ.get("BLACKSHEEP_SECRET_PREFIX", "APP_SECRET")

    for key, value in os.environ.items():
        if key.startswith(env_var_key_prefix) or key.startswith(
            env_var_key_prefix.replace("_", "")
        ):
            app_secrets.append(value)

    if app_secrets:
        return app_secrets

    # For best user experience, here new secrets are generated on the fly.
    logger.debug(
        "Generating secrets on the fly. Configure application secrets to support "
        "tokens validation across restarts and when using multiple instances of the "
        "application!"
    )

    return [generate_secret() for _ in range(3)]


def get_serializer(
    secret_keys: Sequence[str | Secret] | None = None, purpose: str = "dataprotection"
) -> Serializer[str]:
    secrets: list[str]
    if secret_keys:
        secrets = [secret.get_value() for secret in normalize_secrets(secret_keys)]
    if not secret_keys:
        # secrets obtained from env variables or generated on the fly
        secrets = get_keys()
    return URLSafeSerializer(list(secrets), salt=purpose.encode())


def issue_deprecation_warning_for_secret_str():
    warnings.warn(
        "Passing secrets as plain text strings is deprecated and will "
        "be removed in 2.5.x or 2.6.x. Use essentials.secrets.Secret instead.",
        DeprecationWarning,
        stacklevel=2,
    )


def normalize_secrets(secrets: Sequence[str | Secret]) -> Iterable[Secret]:
    """
    Normalize a sequence of secrets into Secret objects.

    Converts plain text strings to Secret objects while preserving existing
    Secret instances. Issues a deprecation warning when plain text strings
    are encountered.

    Args:
        secrets: A sequence of secrets as either plain strings or Secret objects.

    Yields:
        Secret: Normalized Secret objects.

    Raises:
        TypeError: If a secret is neither a string nor a Secret object.

    Warning:
        Passing secrets as plain text strings is deprecated and will be removed
        in a future version. Use essentials.secrets.Secret instead.
    """
    for secret in secrets:
        if isinstance(secret, str):
            issue_deprecation_warning_for_secret_str()
            yield Secret.from_plain_text(secret)
        elif isinstance(secret, Secret):
            yield secret
        else:
            raise TypeError(
                f"Expected secret to be str or Secret, got {type(secret).__name__}"
            )



import secrets
from typing import Any

from blacksheep.cookies import Cookie
from blacksheep.messages import Request, Response
from blacksheep.sessions.abc import Session, SessionStore


class InMemorySessionStore(SessionStore):
    """
    An in-memory implementation of SessionStore for managing user sessions. The session
    ID is transmitted in Cookies, to restore the same session information across
    request-response cycles.

    This session store keeps session data in a Python dictionary, mapping session IDs
    to session data. It is suitable for development and testing environments, but not
    recommended for production use as session data will be lost when the application
    restarts and is not shared across multiple processes or servers.

    Args:
        cookie_name (str): The name of the cookie used to store the session ID.

    Methods:
        load(request): Loads the session associated with the request, or creates a new
            one.
        save(request, response, session): Saves the session data and sets the session
            cookie.
    """

    def __init__(self, cookie_name: str = "session"):
        self._session_cookie_name = cookie_name
        self._sessions: dict[str, Any] = {}

    async def load(self, request: Request) -> Session:
        session_id = request.cookies.get(self._session_cookie_name)
        if session_id and session_id in self._sessions:
            return Session(self._sessions[session_id])
        # Create a new session
        session_id = secrets.token_urlsafe(32)
        session = Session({"id": session_id})
        self._sessions[session_id] = session.to_dict()
        return session

    async def save(
        self, request: Request, response: Response, session: Session
    ) -> None:
        # Use existing session_id or the one generated in load()
        session = request.session
        session_id = session["id"]
        if session_id:
            # Update information in-memory
            self._sessions[session_id] = session.to_dict()
            # Set the session_id cookie in the response
            response.set_cookie(
                Cookie(self._session_cookie_name, session_id, http_only=True, path="/")
            )



"""
This file is used to specify Python extensions, which are used when using Cython.
Extensions are used only if the current runtime is CPython and only if there is not an
environment variable: `BLACKSHEEP_NO_EXTENSIONS=1`.
The logic is to support PyPy. See:
https://github.com/Neoteroi/BlackSheep/issues/539#issuecomment-2888631226
"""

import os
from setuptools import Extension, setup
import platform

COMPILE_ARGS = ["-O2"]

# Check for environment variable to skip extensions
skip_ext = os.environ.get("BLACKSHEEP_NO_EXTENSIONS", "0") == "1"


if platform.python_implementation() == "CPython" and not skip_ext:
    ext_modules = [
        Extension(
            "blacksheep.url",
            ["blacksheep/url.c"],
            extra_compile_args=COMPILE_ARGS,
        ),
        Extension(
            "blacksheep.exceptions",
            ["blacksheep/exceptions.c"],
            extra_compile_args=COMPILE_ARGS,
        ),
        Extension(
            "blacksheep.headers",
            ["blacksheep/headers.c"],
            extra_compile_args=COMPILE_ARGS,
        ),
        Extension(
            "blacksheep.cookies",
            ["blacksheep/cookies.c"],
            extra_compile_args=COMPILE_ARGS,
        ),
        Extension(
            "blacksheep.contents",
            ["blacksheep/contents.c"],
            extra_compile_args=COMPILE_ARGS,
        ),
        Extension(
            "blacksheep.messages",
            ["blacksheep/messages.c"],
            extra_compile_args=COMPILE_ARGS,
        ),
        Extension(
            "blacksheep.scribe",
            ["blacksheep/scribe.c"],
            extra_compile_args=COMPILE_ARGS,
        ),
        Extension(
            "blacksheep.baseapp",
            ["blacksheep/baseapp.c"],
            extra_compile_args=COMPILE_ARGS,
        ),
    ]
else:
    ext_modules = []

setup(ext_modules=ext_modules)



class InvalidOperation(Exception):
    def __init__(self, message: str, inner_exception=None):
        super().__init__(message)
        self.inner_exception = inner_exception


class HTTPException(Exception):
    def __init__(self, status: int, message: str = "HTTP exception"):
        super().__init__(message)
        self.status = status


class BadRequest(HTTPException):
    def __init__(self, message=None):
        super().__init__(400, message or "Bad request")


class BadRequestFormat(BadRequest):
    def __init__(self, message: str, inner_exception=None):
        super().__init__(message)
        self.inner_exception = inner_exception


class FailedRequestError(HTTPException):
    def __init__(self, status: int, data: str) -> None:
        super().__init__(
            status,
            f"The response status code does not indicate success: {status}. "
            "Response body: {data}",
        )
        self.data = data


class NotFound(HTTPException):
    def __init__(self, message=None):
        super().__init__(404, message or "Not found")


class Unauthorized(HTTPException):
    def __init__(self, message=None):
        super().__init__(401, message or "Unauthorized")


class Forbidden(HTTPException):
    def __init__(self, message=None):
        super().__init__(403, message or "Forbidden")


class Conflict(HTTPException):
    def __init__(self, message=None):
        super().__init__(409, message or "Conflict")


class RangeNotSatisfiable(HTTPException):
    def __init__(self):
        super().__init__(416, "Range not satisfiable")


class InternalServerError(HTTPException):
    def __init__(self, source_error: Exception = None):
        super().__init__(500, "Internal server error")
        self.source_error = source_error


class NotImplementedByServer(HTTPException):
    def __init__(self):
        super().__init__(501, "Not implemented by server")


class InvalidArgument(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class MessageAborted(Exception):
    def __init__(self):
        super().__init__(
            "The message was aborted before the client sent its whole content."
        )


class InvalidExceptionHandler(Exception):

    def __init__(self) -> None:
        super().__init__(
            "Invalid Exception handler. Exception handlers must be asynchronous "
            "functions with signature: "
            "async def handler(app: Application, request: Request, exc: Exception)"
        )



import sys
from dataclasses import dataclass
from datetime import datetime
from typing import Any, List, Literal, Sequence, Set, Tuple, Type
from uuid import UUID

import pytest
from guardpost import Identity
from pydantic import BaseModel
from pytest import raises
from rodi import Container

from blacksheep import FormContent, FormPart, JSONContent, MultiPartFormData, Request
from blacksheep.server.bindings import (
    BadRequest,
    Binder,
    BinderAlreadyDefinedException,
    BinderNotRegisteredForValueType,
    BodyBinder,
    BoundValue,
    CookieBinder,
    FormBinder,
    HeaderBinder,
    IdentityBinder,
    InvalidRequestBody,
    JSONBinder,
    MissingBodyError,
    MissingConverterError,
    NameAliasAlreadyDefinedException,
    QueryBinder,
    RequestMethodBinder,
    RequestURLBinder,
    RouteBinder,
    ServiceBinder,
    SyncBinder,
    TypeAliasAlreadyDefinedException,
    get_binder_by_type,
)
from blacksheep.url import URL

JSONContentType = (b"Content-Type", b"application/json")


ExampleLiteralStr = Literal["Hello", "World"]
ExampleLiteralInt = Literal[1, 2, 3]


try:
    # Supported only in Python >= 3.11
    from enum import IntEnum, StrEnum
except ImportError:

    class StrEnum: ...

    class IntEnum: ...


class ExampleOne:
    def __init__(self, a, b):
        self.a = a
        self.b = int(b)


@dataclass
class ExampleDataClass:
    a: str
    b: int


class ExampleTwo:
    def __init__(self, a, b, **kwargs):
        self.a = a
        self.b = b


class ExampleThree:
    def __init__(self, a: str, b: list[str]):
        self.a = a
        self.b = b


class ExampleStrEnum(StrEnum):
    ONE = "one"
    TWO = "two"
    THREE = "three"


class ExampleIntEnum(IntEnum):
    ONE = 1
    TWO = 2
    THREE = 3


class ExamplePydanticModel(BaseModel):
    a: str
    b: int


# Example plain classes


class ContactInfo:
    def __init__(self, phone: str, email: str, created_at: datetime):
        self.phone = phone
        self.email = email
        self.created_at = created_at


class PlainAddress:
    def __init__(self, street: str, city: str, zip_code: str):
        self.street = street
        self.city = city
        self.zip_code = zip_code


class PlainUser:
    def __init__(self, name: str, email: str, age: int, address: PlainAddress):
        self.name = name
        self.email = email
        self.age = age
        self.address = address


class PlainUser2:
    def __init__(self, name: str, email: str, age: int, addresses: list[PlainAddress]):
        self.name = name
        self.email = email
        self.age = age
        self.addresses = addresses


class PlainUserWithContacts:
    def __init__(
        self, name: str, email: str, age: int, contacts: dict[str, ContactInfo]
    ):
        self.name = name
        self.email = email
        self.age = age
        self.contacts = contacts


class PlainUserWithContactsUUID:
    def __init__(
        self, name: str, email: str, age: int, contacts: dict[UUID, ContactInfo]
    ):
        self.name = name
        self.email = email
        self.age = age
        self.contacts = contacts


class ContactInfoModel(BaseModel):
    phone: str
    email: str
    created_at: datetime


class AddressModel(BaseModel):
    street: str
    city: str
    zip_code: str


class UserModel(BaseModel):
    name: str
    email: str
    age: int
    address: AddressModel


class UserModel2(BaseModel):
    name: str
    email: str
    age: int
    addresses: list[AddressModel]


class UserModelWithContacts(BaseModel):
    name: str
    email: str
    age: int
    contacts: dict[str, ContactInfoModel]


class UserModelWithContactsUUID(BaseModel):
    name: str
    email: str
    age: int
    contacts: dict[UUID, ContactInfoModel]


@dataclass
class AddressDc:
    street: str
    city: str
    zip_code: str


@dataclass
class ContactInfoDc:
    phone: str
    email: str
    created_at: datetime


@dataclass
class UserDc:
    name: str
    email: str
    age: int
    address: AddressDc


@dataclass
class UserDcMix:
    name: str
    email: str
    age: int
    address: AddressModel


@dataclass
class UserDcWithContacts:
    name: str
    email: str
    age: int
    contacts: dict[str, ContactInfoDc]


@dataclass
class UserDcWithContactsUUID:
    name: str
    email: str
    age: int
    contacts: dict[UUID, ContactInfoDc]


@dataclass
class UserDc2:
    name: str
    email: str
    age: int
    addresses: list[AddressDc]


@dataclass
class UserDc2Mix:
    name: str
    email: str
    age: int
    addresses: list[AddressModel]  # mix pydantic model and dataclass — weird if done!


async def test_from_body_json_binding():
    request = Request("POST", b"/", [JSONContentType]).with_content(
        JSONContent({"a": "world", "b": 9000})
    )

    parameter = JSONBinder(ExampleOne)

    value = await parameter.get_value(request)

    assert isinstance(value, ExampleOne)
    assert value.a == "world"
    assert value.b == 9000


async def test_from_body_json_binding_extra_parameters_strategy():
    request = Request("POST", b"/", [JSONContentType]).with_content(
        JSONContent(
            {
                "a": "world",
                "b": 9000,
                "c": "This is an extra parameter, accepted by constructor explicitly",
            }
        )
    )

    parameter = JSONBinder(ExampleTwo)

    value = await parameter.get_value(request)

    assert isinstance(value, ExampleTwo)
    assert value.a == "world"
    assert value.b == 9000


async def test_from_body_json_with_converter():
    request = Request("POST", b"/", [JSONContentType]).with_content(
        JSONContent(
            {
                "a": "world",
                "b": 9000,
                "c": "This is an extra parameter, accepted by constructor explicitly",
            }
        )
    )

    def convert(data):
        return ExampleOne(data.get("a"), data.get("b"))

    parameter = JSONBinder(ExampleOne, converter=convert)

    value = await parameter.get_value(request)

    assert isinstance(value, ExampleOne)
    assert value.a == "world"
    assert value.b == 9000


async def test_from_body_json_binding_request_missing_content_type():
    request = Request("POST", b"/", [])

    parameter = JSONBinder(ExampleOne)

    value = await parameter.get_value(request)

    assert value is None


async def test_from_body_json_binding_invalid_input():
    request = Request("POST", b"/", [JSONContentType]).with_content(
        JSONContent({"c": 1, "d": 2})
    )

    parameter = JSONBinder(ExampleOne)

    with raises(BadRequest):
        await parameter.get_value(request)


@pytest.mark.parametrize(
    "expected_type,header_value,expected_value",
    [
        [str, b"Foo", "Foo"],
        [str, b"foo", "foo"],
        [str, b"Hello%20World%21%3F", "Hello World!?"],
        [int, b"1", 1],
        [int, b"10", 10],
        [float, b"1.5", 1.5],
        [float, b"1241.5", 1241.5],
        [bool, b"1", True],
        [bool, b"0", False],
        [ExampleLiteralStr, b"Hello", "Hello"],
        [ExampleLiteralInt, b"2", 2],
    ],
)
async def test_from_header_binding(expected_type, header_value, expected_value):
    request = Request("GET", b"/", [(b"X-Foo", header_value)])

    parameter = HeaderBinder(expected_type, "X-Foo")

    value = await parameter.get_value(request)

    # the following assertion does not apply to Literal
    if expected_type not in {ExampleLiteralStr, ExampleLiteralInt}:
        assert isinstance(value, expected_type)
    assert value == expected_value


# TODO: merge with above when support for Python < 3.11 is dropped
@pytest.mark.skipif(sys.version_info < (3, 11), reason="requires python3.11 or higher")
@pytest.mark.parametrize(
    "expected_type,header_value,expected_value",
    [
        [ExampleStrEnum, b"one", ExampleStrEnum.ONE],
        [ExampleStrEnum, b"ONE", ExampleStrEnum.ONE],
        [ExampleIntEnum, b"1", ExampleIntEnum.ONE],
        [ExampleIntEnum, b"ONE", ExampleIntEnum.ONE],
    ],
)
async def test_from_header_binding_enums(expected_type, header_value, expected_value):
    await test_from_header_binding(expected_type, header_value, expected_value)


@pytest.mark.parametrize(
    "expected_type,header_value,expected_value",
    [
        [str, b"Foo", "Foo"],
        [str, b"foo", "foo"],
        [str, b"\xc5\x81ukasz", "Łukasz"],
        [str, b"Hello%20World%21%3F", "Hello World!?"],
        [int, b"1", 1],
        [int, b"10", 10],
        [float, b"1.5", 1.5],
        [float, b"1241.5", 1241.5],
        [bool, b"1", True],
        [bool, b"0", False],
    ],
)
async def test_from_header_binding_name_ci(expected_type, header_value, expected_value):
    request = Request("GET", b"/", [(b"X-Foo", header_value)])

    parameter = HeaderBinder(expected_type, "x-foo")

    value = await parameter.get_value(request)

    assert isinstance(value, expected_type)
    assert value == expected_value


@pytest.mark.parametrize(
    "expected_type,query_value,expected_value",
    [
        [str, b"Foo", "Foo"],
        [str, b"foo", "foo"],
        [str, b"Hello%20World%21%3F", "Hello World!?"],
        [int, b"1", 1],
        [int, b"10", 10],
        [float, b"1.5", 1.5],
        [float, b"1241.5", 1241.5],
        [bool, b"1", True],
        [bool, b"0", False],
        [ExampleLiteralStr, b"Hello", "Hello"],
        [ExampleLiteralInt, b"2", 2],
    ],
)
async def test_from_query_binding(expected_type, query_value, expected_value):
    request = Request("GET", b"/?foo=" + query_value, None)

    parameter = QueryBinder(expected_type, "foo")

    value = await parameter.get_value(request)

    # assertion not applicable to Literal
    if expected_type not in {ExampleLiteralStr, ExampleLiteralInt}:
        assert isinstance(value, expected_type)
    assert value == expected_value


# TODO: merge with above when support for Python < 3.11 is dropped
@pytest.mark.skipif(sys.version_info < (3, 11), reason="requires python3.11 or higher")
@pytest.mark.parametrize(
    "expected_type,query_value,expected_value",
    [
        [ExampleStrEnum, b"one", ExampleStrEnum.ONE],
        [ExampleIntEnum, b"1", ExampleIntEnum.ONE],
    ],
)
async def test_from_query_binding_enums(expected_type, query_value, expected_value):
    await test_from_query_binding(expected_type, query_value, expected_value)


@pytest.mark.parametrize(
    "expected_type,route_value,expected_value",
    [
        [str, "Foo", "Foo"],
        [str, "foo", "foo"],
        [str, "Hello%20World%21%3F", "Hello World!?"],
        [int, "1", 1],
        [int, "10", 10],
        [float, "1.5", 1.5],
        [float, "1241.5", 1241.5],
        [bool, "1", True],
        [bool, "0", False],
        [
            UUID,
            "b0c1f822-b63c-475e-9f2e-b6406bafcc2b",
            UUID("b0c1f822-b63c-475e-9f2e-b6406bafcc2b"),
        ],
    ],
)
async def test_from_route_binding(expected_type, route_value, expected_value):
    request = Request("GET", b"/", None)
    request.route_values = {"name": route_value}

    parameter = RouteBinder(expected_type, "name")

    value = await parameter.get_value(request)

    if expected_type not in {ExampleLiteralStr, ExampleLiteralInt}:
        assert isinstance(value, expected_type)
    assert value == expected_value


# TODO: merge with above when support for Python < 3.11 is dropped
@pytest.mark.skipif(sys.version_info < (3, 11), reason="requires python3.11 or higher")
@pytest.mark.parametrize(
    "expected_type,route_value,expected_value",
    [
        [ExampleStrEnum, "one", ExampleStrEnum.ONE],
        [ExampleIntEnum, "1", ExampleIntEnum.ONE],
    ],
)
async def test_from_route_binding_enums(expected_type, route_value, expected_value):
    await test_from_route_binding(expected_type, route_value, expected_value)


@pytest.mark.parametrize("binder_type", [HeaderBinder, QueryBinder, RouteBinder])
async def test_raises_for_missing_default_converter(binder_type):
    with raises(MissingConverterError):
        binder_type("example", ExampleOne)


@pytest.mark.parametrize(
    "expected_type,invalid_value",
    [[int, "x"], [int, ""], [float, "x"], [float, ""], [bool, "x"], [bool, "yes"]],
)
async def test_from_route_raises_for_invalid_parameter(expected_type, invalid_value):
    request = Request("GET", b"/", None)
    request.route_values = {"name": invalid_value}

    parameter = RouteBinder(expected_type, "name")

    with raises(BadRequest):
        await parameter.get_value(request)


@pytest.mark.parametrize(
    "expected_type,invalid_value",
    [
        [int, b"x"],
        [int, b""],
        [float, b"x"],
        [float, b""],
        [bool, b"x"],
        [bool, b"yes"],
    ],
)
async def test_from_query_raises_for_invalid_parameter(
    expected_type, invalid_value: bytes
):
    request = Request("GET", b"/?foo=" + invalid_value, None)

    parameter = QueryBinder(expected_type, "foo", required=True)

    with raises(BadRequest):
        await parameter.get_value(request)


async def test_from_services():
    request = Request("GET", b"/", [])

    service_instance = ExampleOne(1, 2)
    container = Container()
    container.add_instance(service_instance)

    parameter = ServiceBinder(ExampleOne, "service", False, container)
    value = await parameter.get_value(request)

    assert value is service_instance


@pytest.mark.parametrize(
    "declared_type,expected_type,header_values,expected_values",
    [
        [list[str], list, [b"Lorem", b"ipsum", b"dolor"], ["Lorem", "ipsum", "dolor"]],
        [
            tuple[str],
            tuple,
            [b"Lorem", b"ipsum", b"dolor"],
            ("Lorem", "ipsum", "dolor"),
        ],
        [Set[str], set, [b"Lorem", b"ipsum", b"dolor"], {"Lorem", "ipsum", "dolor"}],
        [
            Sequence[str],
            list,
            [b"Lorem", b"ipsum", b"dolor"],
            ["Lorem", "ipsum", "dolor"],
        ],
    ],
)
async def test_from_header_binding_iterables(
    declared_type, expected_type, header_values, expected_values
):
    request = Request("GET", b"/", [(b"X-Foo", value) for value in header_values])

    parameter = HeaderBinder(declared_type, "X-Foo")

    value = await parameter.get_value(request)

    assert isinstance(value, expected_type)
    assert value == expected_values


@pytest.mark.parametrize(
    "declared_type,expected_type,query_values,expected_values",
    [
        [list, list, [b"Lorem", b"ipsum", b"dolor"], ["Lorem", "ipsum", "dolor"]],
        [tuple, tuple, [b"Lorem", b"ipsum", b"dolor"], ("Lorem", "ipsum", "dolor")],
        [set, set, [b"Lorem", b"ipsum", b"dolor"], {"Lorem", "ipsum", "dolor"}],
        [List, list, [b"Lorem", b"ipsum", b"dolor"], ["Lorem", "ipsum", "dolor"]],
        [Tuple, tuple, [b"Lorem", b"ipsum", b"dolor"], ("Lorem", "ipsum", "dolor")],
        [Set, set, [b"Lorem", b"ipsum", b"dolor"], {"Lorem", "ipsum", "dolor"}],
        [list[str], list, [b"Lorem", b"ipsum", b"dolor"], ["Lorem", "ipsum", "dolor"]],
        [
            tuple[str],
            tuple,
            [b"Lorem", b"ipsum", b"dolor"],
            ("Lorem", "ipsum", "dolor"),
        ],
        [Set[str], set, [b"Lorem", b"ipsum", b"dolor"], {"Lorem", "ipsum", "dolor"}],
        [
            Sequence[str],
            list,
            [b"Lorem", b"ipsum", b"dolor"],
            ["Lorem", "ipsum", "dolor"],
        ],
        [list[int], list, [b"10"], [10]],
        [list[int], list, [b"0", b"1", b"0"], [0, 1, 0]],
        [list[int], list, [b"0", b"1", b"0", b"2"], [0, 1, 0, 2]],
        [list[int], list, [b"0", b"1", b"0", b"2"], [0, 1, 0, 2]],
        [list[bytes], list, [b"0", b"1", b"0", b"2"], [b"0", b"1", b"0", b"2"]],
        [list[bool], list, [b"1"], [True]],
        [list[bool], list, [b"0", b"1", b"0"], [False, True, False]],
        [list[bool], list, [b"0", b"1", b"0", b"true"], [False, True, False, True]],
        [list[bool], list, [b"0", b"1", b"0", b"true"], [False, True, False, True]],
        [list[float], list, [b"10.2"], [10.2]],
        [list[float], list, [b"0.3", b"1", b"0"], [0.3, 1.0, 0]],
        [list[float], list, [b"0.5", b"1", b"0", b"2"], [0.5, 1.0, 0, 2.0]],
        [list[float], list, [b"0.5", b"1", b"0", b"2"], [0.5, 1.0, 0, 2.0]],
        [tuple[float], tuple, [b"10.2"], (10.2,)],
        [tuple[float], tuple, [b"0.3", b"1", b"0"], (0.3, 1.0, 0)],
        [tuple[float], tuple, [b"0.5", b"1", b"0", b"2"], (0.5, 1.0, 0, 2.0)],
        [tuple[float], tuple, [b"0.5", b"1", b"0", b"2"], (0.5, 1.0, 0, 2.0)],
        [Set[int], set, [b"10"], {10}],
        [Set[int], set, [b"0", b"1", b"0"], {0, 1, 0}],
        [Set[int], set, [b"0", b"1", b"0", b"2"], {0, 1, 0, 2}],
        [set[int], set, [b"0", b"1", b"0", b"2"], {0, 1, 0, 2}],
        [
            list[UUID],
            list,
            [
                b"de18d268-f5c5-42db-89b2-c61bbfe96e65",
                b"d5fd0cde-4ad6-4b61-a5b1-5b8e6d48cebe",
                b"00000000-0000-0000-0000-000000000000",
            ],
            [
                UUID("de18d268-f5c5-42db-89b2-c61bbfe96e65"),
                UUID("d5fd0cde-4ad6-4b61-a5b1-5b8e6d48cebe"),
                UUID("00000000-0000-0000-0000-000000000000"),
            ],
        ],
        [
            list[UUID],
            list,
            [
                b"de18d268-f5c5-42db-89b2-c61bbfe96e65",
                b"d5fd0cde-4ad6-4b61-a5b1-5b8e6d48cebe",
                b"00000000-0000-0000-0000-000000000000",
            ],
            [
                UUID("de18d268-f5c5-42db-89b2-c61bbfe96e65"),
                UUID("d5fd0cde-4ad6-4b61-a5b1-5b8e6d48cebe"),
                UUID("00000000-0000-0000-0000-000000000000"),
            ],
        ],
    ],
)
async def test_from_query_binding_iterables(
    declared_type, expected_type, query_values, expected_values
):
    qs = b"&foo=".join([value for value in query_values])

    request = Request("GET", b"/?foo=" + qs, None)

    parameter = QueryBinder(declared_type, "foo")

    values = await parameter.get_value(request)

    assert isinstance(values, expected_type)
    assert values == expected_values


@pytest.mark.parametrize(
    "declared_type", [list[list[str]], tuple[tuple[str]], list[list]]
)
async def test_nested_iterables_raise_missing_converter_from_header(declared_type):
    with raises(MissingConverterError):
        HeaderBinder(declared_type)


@pytest.mark.parametrize(
    "declared_type", [list[list[str]], tuple[tuple[str]], list[list]]
)
async def test_nested_iterables_raise_missing_converter_from_query(declared_type):
    with raises(MissingConverterError):
        QueryBinder("example", declared_type)


async def test_identity_binder_identity_not_set():
    request = Request("GET", b"/", None)

    parameter = IdentityBinder(Identity)

    value = await parameter.get_value(request)

    # request.user is automatically set, if missing, to an empty object representing
    # an anonymous user
    assert value is not None


async def test_identity_binder():
    request = Request("GET", b"/", None)
    request.user = Identity()

    parameter = IdentityBinder(Identity)

    value = await parameter.get_value(request)

    assert value is request.user


async def test_from_body_form_binding_urlencoded():
    request = Request("POST", b"/", []).with_content(
        FormContent({"a": "world", "b": 9000})
    )

    parameter = FormBinder(ExampleOne)

    value = await parameter.get_value(request)

    assert isinstance(value, ExampleOne)
    assert value.a == "world"
    assert value.b == 9000


async def test_from_body_form_binding_urlencoded_keys_duplicates():
    request = Request("POST", b"/", []).with_content(
        FormContent([("a", "world"), ("b", "one"), ("b", "two"), ("b", "three")])
    )

    parameter = FormBinder(ExampleThree)

    value = await parameter.get_value(request)

    assert isinstance(value, ExampleThree)
    assert value.a == "world"
    assert value.b == ["one", "two", "three"]


async def test_from_body_form_binding_multipart():
    request = Request("POST", b"/", []).with_content(
        MultiPartFormData([FormPart(b"a", b"world"), FormPart(b"b", b"9000")])
    )

    parameter = FormBinder(ExampleOne)
    value = await parameter.get_value(request)

    assert isinstance(value, ExampleOne)
    assert value.a == "world"
    assert value.b == 9000


async def test_from_body_form_binding_multipart_keys_duplicates():
    request = Request("POST", b"/", []).with_content(
        MultiPartFormData(
            [
                FormPart(b"a", b"world"),
                FormPart(b"b", b"one"),
                FormPart(b"b", b"two"),
                FormPart(b"b", b"three"),
            ]
        )
    )

    parameter = FormBinder(ExampleThree)

    value = await parameter.get_value(request)

    assert isinstance(value, ExampleThree)
    assert value.a == "world"
    assert value.b == ["one", "two", "three"]


async def test_custom_bound_value_and_binder():
    class FromMethod(BoundValue[str]):
        pass

    class MethodBinder(Binder):
        handle = FromMethod

        async def get_value(self, request: Request) -> str | None:
            return request.method

    parameter = MethodBinder(str)

    for method in {"GET", "POST", "TRACE"}:
        value = await parameter.get_value(Request(method, b"/", []))
        assert value == method


async def test_custom_bound_value_fails_for_missing_binder():
    class FromSomething(BoundValue[str]):
        pass

    def faulty(example: FromSomething):
        pass

    with pytest.raises(BinderNotRegisteredForValueType):
        get_binder_by_type(FromSomething)


async def test_raises_for_duplicate_binders():
    class FromMethod(BoundValue[str]):
        pass

    class MethodBinder(Binder):
        handle = FromMethod

        async def get_value(self, request: Request) -> str | None:
            return request.method

    with pytest.raises(BinderAlreadyDefinedException):

        class MethodBinder2(Binder):
            handle = FromMethod

            async def get_value(self, request: Request) -> str | None:
                return request.method


@pytest.mark.parametrize(
    "binder_type,expected_source_name",
    [(RouteBinder, "route"), (QueryBinder, "query"), (HeaderBinder, "header")],
)
def test_sync_binder_source_name(
    binder_type: Type[SyncBinder], expected_source_name: str
):
    binder = binder_type(str)
    assert binder.source_name == expected_source_name


async def test_body_binder_throws_for_abstract_methods():
    body_binder = BodyBinder(dict)

    with pytest.raises(NotImplementedError):
        body_binder.matches_content_type(Request("HEAD", b"/", []))

    with pytest.raises(NotImplementedError):
        await body_binder.read_data(Request("HEAD", b"/", []))


async def test_body_binder_throws_bad_request_for_missing_body():
    class CustomBodyBinder(BodyBinder):
        def matches_content_type(self, request: Request) -> bool:
            return True

        async def read_data(self, request: Request) -> Any:
            return None

    body_binder = CustomBodyBinder(dict)

    with pytest.raises(MissingBodyError):
        await body_binder.get_value(
            Request("POST", b"/", [(b"content-type", b"application/json")])
        )

    body_binder = JSONBinder(dict, required=True)

    with pytest.raises(MissingBodyError):
        await body_binder.get_value(Request("POST", b"/", []))


async def test_body_binder_throws_bad_request_for_value_error():
    body_binder = JSONBinder(dict, required=True)

    def example_converter(value):
        raise ValueError("Invalid value")

    body_binder.converter = example_converter

    with pytest.raises(InvalidRequestBody):
        await body_binder.get_value(
            Request(
                "POST", b"/", [(b"content-type", b"application/json")]
            ).with_content(JSONContent({"id": "1", "name": "foo"}))
        )


def test_sync_binders_source_name():
    assert CookieBinder().source_name == "cookie"
    assert HeaderBinder().source_name == "header"
    assert QueryBinder().source_name == "query"
    assert RouteBinder().source_name == "route"


@pytest.mark.parametrize("method", ["GET", "OPTIONS", "POST"])
async def test_request_method_binder(method):
    request = Request(method, b"/", [])
    parameter = RequestMethodBinder()
    value = await parameter.get_value(request)
    assert value == method


@pytest.mark.parametrize("url", [b"/", b"/api/cats/123", b"/foo/index.html?s=20"])
async def test_request_url_binder(url):
    request = Request("GET", url, [])
    parameter = RequestURLBinder()
    value = await parameter.get_value(request)
    assert value == URL(url)


def test_duplicate_name_alias_raises():
    class FooBinder1(Binder):
        name_alias = "foo_absurd_example"

    with pytest.raises(NameAliasAlreadyDefinedException) as duplicate_alias_error:

        class FooBinder2(Binder):
            name_alias = "foo_absurd_example"

    assert str(duplicate_alias_error.value) == (
        "There is already a name alias defined for 'foo_absurd_example', "
        "the second type is: FooBinder2"
    )


def test_duplicate_type_alias_raises():
    class X:
        pass

    class XBinder1(Binder):
        type_alias = X

    with pytest.raises(TypeAliasAlreadyDefinedException) as duplicate_alias_error:

        class XBinder2(Binder):
            type_alias = X

    assert str(duplicate_alias_error.value) == (
        "There is already a type alias defined for 'X', the second type is: XBinder2"
    )


@pytest.mark.parametrize(
    "model_class,expected_a,expected_b",
    [
        (ExampleOne, "world", 9000),  # Plain class
        (ExampleDataClass, "world", 9000),  # Dataclass
        pytest.param(
            ExamplePydanticModel,
            "world",
            9000,  # Pydantic model
            marks=pytest.mark.skipif(
                BaseModel is None, reason="pydantic not available"
            ),
        ),
    ],
)
async def test_from_body_json_binding_ignore_extra_parameters(
    model_class, expected_a, expected_b
):
    request = Request("POST", b"/", [JSONContentType]).with_content(
        JSONContent(
            {
                "a": "world",
                "b": 9000,
                "c": "This is an extra parameter that should be ignored",
            }
        )
    )

    parameter = JSONBinder(model_class)

    value = await parameter.get_value(request)

    assert isinstance(value, model_class)
    assert value.a == expected_a
    assert value.b == expected_b


@pytest.mark.parametrize("expected_type", [PlainUser, UserModel, UserDc, UserDcMix])
async def test_from_body_json_binding_ignore_extra_parameters_nested_1(expected_type):
    # Test that extra properties are ignored also in child properties
    plain_data = {
        "name": "Jane",
        "email": "jane@example.com",
        "age": 25,
        "extra_field": "ignored",
        "address": {
            "street": "456 Oak Ave",
            "city": "Seattle",
            "zip_code": "98101",
            "extra_address_field": "ignored",
        },
    }

    request = Request("POST", b"/", [JSONContentType]).with_content(
        JSONContent(plain_data)
    )

    parameter = JSONBinder(expected_type)

    value = await parameter.get_value(request)

    assert isinstance(value, expected_type)
    assert value.name == "Jane"
    assert value.address.street == "456 Oak Ave"


@pytest.mark.parametrize("expected_type", [PlainUser2, UserModel2, UserDc2, UserDc2Mix])
async def test_from_body_json_binding_ignore_extra_parameters_nested_2(expected_type):
    # Test that extra properties are ignored also in child properties
    plain_data = {
        "name": "Jane",
        "email": "jane@example.com",
        "age": 25,
        "extra_field": "ignored",
        "addresses": [
            {
                "street": "456 Oak Ave",
                "city": "Seattle",
                "zip_code": "98101",
                "extra_address_field": "ignored",
            },
            {
                "street": "Foo",
                "city": "Foo City",
                "zip_code": "00888",
                "extra_address_field": "ignored",
                "some_other_field": 3,
            },
        ],
    }

    request = Request("POST", b"/", [JSONContentType]).with_content(
        JSONContent(plain_data)
    )

    parameter = JSONBinder(expected_type)

    value = await parameter.get_value(request)

    assert isinstance(value, expected_type)
    assert value.name == "Jane"
    assert value.addresses[0].street == "456 Oak Ave"
    assert value.addresses[1].street == "Foo"


@pytest.mark.parametrize(
    "collection_type,model_class,expected_type",
    [
        (list[ExampleOne], ExampleOne, list),
        (list[ExampleDataClass], ExampleDataClass, list),
        pytest.param(
            list[ExamplePydanticModel],
            ExamplePydanticModel,
            list,
            marks=pytest.mark.skipif(
                BaseModel is None, reason="pydantic not available"
            ),
        ),
        (Sequence[ExampleOne], ExampleOne, list),
        (Sequence[ExampleDataClass], ExampleDataClass, list),
        pytest.param(
            Sequence[ExamplePydanticModel],
            ExamplePydanticModel,
            list,
            marks=pytest.mark.skipif(
                BaseModel is None, reason="pydantic not available"
            ),
        ),
    ],
)
async def test_from_body_json_binding_collections(
    collection_type, model_class, expected_type
):
    request = Request("POST", b"/", [JSONContentType]).with_content(
        JSONContent(
            [
                {"a": "first", "b": 100, "c": "extra property to ignore 1"},
                {"a": "second", "b": 200, "c": "extra property to ignore 2"},
                {"a": "third", "b": 300, "c": "extra property to ignore 3"},
            ]
        )
    )

    parameter = JSONBinder(collection_type)

    value = await parameter.get_value(request)

    assert isinstance(value, expected_type)
    assert len(value) == 3

    # Convert to list for uniform iteration (since sets don't guarantee order)
    items = list(value)

    for i, item in enumerate(items):
        assert isinstance(item, model_class)
        # For sets, we can't guarantee order, so we check that all expected values are present
        if expected_type == set:
            assert item.a in ["first", "second", "third"]
            assert item.b in [100, 200, 300]
        else:
            # For lists and sequences, order is preserved
            expected_values = [("first", 100), ("second", 200), ("third", 300)]
            assert item.a == expected_values[i][0]
            assert item.b == expected_values[i][1]


@pytest.mark.parametrize(
    "expected_type",
    [
        PlainUserWithContactsUUID,
        UserModelWithContactsUUID,
        UserDcWithContactsUUID,
    ],
)
async def test_from_body_json_binding_dict_uuid_custom_class(expected_type):
    """Test conversion of dict[UUID, CustomClass] from JSON body"""
    uuid_home = "b0c1f822-b63c-475e-9f2e-b6406bafcc2b"
    uuid_work = "d5fd0cde-4ad6-4b61-a5b1-5b8e6d48cebe"
    uuid_mobile = "00000000-0000-0000-0000-000000000000"

    data = {
        "name": "Jane",
        "email": "jane@example.com",
        "age": 25,
        "contacts": {
            uuid_home: {
                "phone": "555-1234",
                "email": "jane.home@example.com",
                "created_at": "2023-01-15T10:30:00",
                "extra_field": "should be ignored",
            },
            uuid_work: {
                "phone": "555-5678",
                "email": "jane.work@example.com",
                "created_at": "2023-02-20T14:45:00",
                "extra_field": "should be ignored",
            },
            uuid_mobile: {
                "phone": "555-9999",
                "email": "jane.mobile@example.com",
                "created_at": "2023-03-25T08:15:00",
                "extra_field": "should be ignored",
            },
        },
    }

    request = Request("POST", b"/", [JSONContentType]).with_content(JSONContent(data))

    parameter = JSONBinder(expected_type)

    value = await parameter.get_value(request)

    assert isinstance(value, expected_type)
    assert value.name == "Jane"
    assert value.email == "jane@example.com"
    assert value.age == 25
    assert isinstance(value.contacts, dict)
    assert len(value.contacts) == 3

    # Check that keys are UUIDs
    for key in value.contacts.keys():
        assert isinstance(key, UUID)

    # Check specific values
    assert UUID(uuid_home) in value.contacts
    assert UUID(uuid_work) in value.contacts
    assert UUID(uuid_mobile) in value.contacts
    assert value.contacts[UUID(uuid_home)].phone == "555-1234"
    assert value.contacts[UUID(uuid_home)].email == "jane.home@example.com"
    assert value.contacts[UUID(uuid_home)].created_at == datetime(
        2023, 1, 15, 10, 30, 0
    )
    assert value.contacts[UUID(uuid_work)].phone == "555-5678"
    assert value.contacts[UUID(uuid_work)].email == "jane.work@example.com"
    assert value.contacts[UUID(uuid_work)].created_at == datetime(
        2023, 2, 20, 14, 45, 0
    )
    assert value.contacts[UUID(uuid_mobile)].created_at == datetime(
        2023, 3, 25, 8, 15, 0
    )


@pytest.mark.parametrize(
    "expected_type",
    [
        PlainUserWithContacts,
        UserModelWithContacts,
        UserDcWithContacts,
    ],
)
async def test_from_body_json_binding_dict_str_custom_class(expected_type):
    """Test conversion of dict[str, CustomClass] from JSON body"""
    data = {
        "name": "Jane",
        "email": "jane@example.com",
        "age": 25,
        "contacts": {
            "home": {
                "phone": "555-1234",
                "email": "jane.home@example.com",
                "created_at": "2023-01-15T10:30:00",
            },
            "work": {
                "phone": "555-5678",
                "email": "jane.work@example.com",
                "created_at": "2023-02-20T14:45:00",
            },
            "mobile": {
                "phone": "555-9999",
                "email": "jane.mobile@example.com",
                "created_at": "2023-03-25T08:15:00",
            },
        },
    }

    request = Request("POST", b"/", [JSONContentType]).with_content(JSONContent(data))

    parameter = JSONBinder(expected_type)

    value = await parameter.get_value(request)

    assert isinstance(value, expected_type)
    assert value.name == "Jane"
    assert value.email == "jane@example.com"
    assert value.age == 25
    assert isinstance(value.contacts, dict)
    assert len(value.contacts) == 3
    assert "home" in value.contacts
    assert "work" in value.contacts
    assert "mobile" in value.contacts
    assert value.contacts["home"].phone == "555-1234"
    assert value.contacts["home"].email == "jane.home@example.com"
    assert value.contacts["home"].created_at == datetime(2023, 1, 15, 10, 30, 0)
    assert value.contacts["work"].phone == "555-5678"
    assert value.contacts["work"].email == "jane.work@example.com"
    assert value.contacts["work"].created_at == datetime(2023, 2, 20, 14, 45, 0)


@pytest.mark.parametrize(
    "expected_type",
    [
        PlainUserWithContactsUUID,
        UserModelWithContactsUUID,
        UserDcWithContactsUUID,
    ],
)
async def test_from_body_json_binding_empty_dict(expected_type):
    """Test conversion with empty dict"""
    data = {
        "name": "Jane",
        "email": "jane@example.com",
        "age": 25,
        "contacts": {},
    }

    request = Request("POST", b"/", [JSONContentType]).with_content(JSONContent(data))

    parameter = JSONBinder(expected_type)

    value = await parameter.get_value(request)

    assert isinstance(value, expected_type)
    assert value.name == "Jane"
    assert value.email == "jane@example.com"
    assert value.age == 25
    assert isinstance(value.contacts, dict)
    assert len(value.contacts) == 0



try:
    import httptools
except ImportError:
    httptools = None


try:
    import h11
except ImportError:
    h11 = None


if httptools is not None:

    class HTTPToolsResponseParser:
        def __init__(self, connection) -> None:
            self.connection = connection
            self._parser = httptools.HttpResponseParser(connection)

        def feed_data(self, data: bytes) -> None:
            self._parser.feed_data(data)

        def get_status_code(self) -> int:
            return self._parser.get_status_code()

        def reset(self) -> None:
            self._parser = httptools.HttpResponseParser(self.connection)


if h11 is not None:

    class H11ResponseParser:
        def __init__(self, connection) -> None:
            self.connection = connection
            self._conn = h11.Connection(h11.CLIENT)
            self._status_code = None
            self._headers = []
            self._complete = False

        def feed_data(self, data: bytes) -> None:
            self._conn.receive_data(data)
            while True:
                event = self._conn.next_event()
                if event is h11.NEED_DATA:
                    break
                if isinstance(event, h11.Response):
                    self._status_code = event.status_code
                    self._headers = [(k, v) for k, v in event.headers]
                    self.connection.headers = self._headers
                    self.connection.on_headers_complete()
                elif isinstance(event, h11.Data):
                    self.connection.on_body(event.data)
                elif isinstance(event, h11.EndOfMessage):
                    self._complete = True
                    self.connection.on_message_complete()
                # Ignore other events

        def get_status_code(self) -> int:
            return self._status_code or 0

        def reset(self) -> None:
            self._conn = h11.Connection(h11.CLIENT)
            self._status_code = None
            self._headers = []
            self._complete = False


def get_default_parser(client_connection):

    if h11 is not None:
        return H11ResponseParser(client_connection)

    if httptools is not None:
        return HTTPToolsResponseParser(client_connection)

    raise RuntimeError(
        "Missing Python dependencies to provide a default HTTP Response parser. "
        "Install either h11 or httptools."
    )
